# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List, Tuple, Union, TYPE_CHECKING, Optional, Any, Dict, Callable
from typing_extensions import Literal

if TYPE_CHECKING:
    from logging import Logger
    from ..._configuration import Configuration
    from .._producer_async import EventHubProducer
    from .._consumer_async import EventHubConsumer
    from ..._common import EventData
    from .._consumer_client_async import EventHubConsumerClient
    from .._client_base_async import ClientBase
    from ..._client_base import _Address
    from ...amqp._amqp_message import AmqpAnnotatedMessage

    try:
        from uamqp.types import AMQPSymbol as uamqp_Types_AMQPSymbol
        from uamqp.constants import (
            ConnectionState as uamqp_ConnectionState,
        )
        from uamqp import (
            Message as uamqp_Message,
            BatchMessage as uamqp_BatchMessage,
            AMQPClientAsync as uamqp_AMQPClient,
            Connection as uamqp_Connection,
            Source as uamqp_Source,
            ReceiveClientAsync as uamqp_ReceiveClient,
            SendClientAsync as uamqp_SendClient,
        )
        from uamqp.authentication import JWTTokenAsync as uamqp_JWTTokenAuth

    except ImportError:
        pass

    from ..._pyamqp.aio import (
        AMQPClientAsync as pyamqp_AMQPClient,
        Connection as pyamqp_Connection,
        ReceiveClientAsync as pyamqp_ReceiveClient,
        SendClientAsync as pyamqp_SendClient,
    )
    from ..._pyamqp.message import (
        Message as pyamqp_Message,
        BatchMessage as pyamqp_BatchMessage,
    )
    from ..._pyamqp.endpoints import Source as pyamqp_Source
    from ..._pyamqp.aio._authentication_async import JWTTokenAuthAsync as pyamqp_JWTTokenAuth
    from ..._pyamqp.constants import (
        ConnectionState as pyamqp_ConnectionState,
    )


class AmqpTransportAsync(ABC):  # pylint: disable=too-many-public-methods
    """
    Abstract class that defines a set of common methods needed by producer and consumer.
    """

    KIND: str

    # define constants
    MAX_FRAME_SIZE_BYTES: int
    MAX_MESSAGE_LENGTH_BYTES: int
    TIMEOUT_FACTOR: int
    CONNECTION_CLOSING_STATES: Tuple[
        Union["uamqp_ConnectionState", "pyamqp_ConnectionState"],
        Union["uamqp_ConnectionState", "pyamqp_ConnectionState"],
        Union["uamqp_ConnectionState", "pyamqp_ConnectionState"],
        Union["uamqp_ConnectionState", "pyamqp_ConnectionState"],
        Optional[Union["uamqp_ConnectionState", "pyamqp_ConnectionState"]],
    ]
    # define symbols
    PRODUCT_SYMBOL: Union[uamqp_Types_AMQPSymbol, Literal["product"]]
    VERSION_SYMBOL: Union[uamqp_Types_AMQPSymbol, Literal["version"]]
    FRAMEWORK_SYMBOL: Union[uamqp_Types_AMQPSymbol, Literal["framework"]]
    PLATFORM_SYMBOL: Union[uamqp_Types_AMQPSymbol, Literal["platform"]]
    USER_AGENT_SYMBOL: Union[uamqp_Types_AMQPSymbol, Literal["user-agent"]]
    PROP_PARTITION_KEY_AMQP_SYMBOL: Union[uamqp_Types_AMQPSymbol, Literal[b"x-opt-partition-key"]]

    @staticmethod
    @abstractmethod
    def build_message(**kwargs: Any) -> Union["uamqp_Message", "pyamqp_Message"]:
        """
        Creates a uamqp.Message or pyamqp.Message with given arguments.
        :rtype: ~uamqp.Message or ~pyamqp.message.Message
        """

    @staticmethod
    @abstractmethod
    def build_batch_message(**kwargs: Any) -> Union["uamqp_BatchMessage", "pyamqp_BatchMessage"]:
        """
        Creates a uamqp.BatchMessage or pyamqp.BatchMessage with given arguments.
        :rtype: ~uamqp.BatchMessage or ~pyamqp.message.BatchMessage
        """

    @staticmethod
    @abstractmethod
    def to_outgoing_amqp_message(annotated_message: AmqpAnnotatedMessage) -> Union["uamqp_Message", "pyamqp_Message"]:
        """
        Converts an AmqpAnnotatedMessage into an Amqp Message.
        :param ~azure.eventhub.amqp.AmqpAnnotatedMessage annotated_message: AmqpAnnotatedMessage to convert.
        :rtype: ~uamqp.Message or ~pyamqp.message.Message
        """

    @staticmethod
    @abstractmethod
    def get_batch_message_encoded_size(message: Union["uamqp_Message", "pyamqp_Message"]) -> int:
        """
        Gets the batch message encoded size given an underlying Message.
        :param ~uamqp.BatchMessage or ~pyamqp.message.Message message: Message to get encoded size of.
        :rtype: int
        """

    @staticmethod
    @abstractmethod
    def get_remote_max_message_size(handler: Union["uamqp_AMQPClient", "pyamqp_AMQPClient"]) -> int:
        """
        Returns max peer message size.
        :param handler: Client to get remote max message size on link from.
        :type handler: ~uamqp.AMQPClientAsync or ~pyamqp.aio.AMQPClientAsync
        :rtype: int
        """

    @staticmethod
    @abstractmethod
    def create_retry_policy(config: Configuration) -> None:
        """
        Creates the error retry policy.
        :param ~azure.eventhub._configuration.Configuration config: Configuration.
        """

    @staticmethod
    @abstractmethod
    def create_link_properties(link_properties: Dict[bytes, int]) -> Dict[bytes, Any]:
        """
        Creates and returns the link properties.
        :param dict[bytes, int] link_properties: The dict of symbols and corresponding values.
        :rtype: dict
        """

    @staticmethod
    @abstractmethod
    async def create_connection_async(
        *,
        endpoint: str,
        auth: Union["uamqp_JWTTokenAuth", pyamqp_JWTTokenAuth],
        container_id: Optional[str] = None,
        max_frame_size: int,
        channel_max: int,
        idle_timeout: Optional[float],
        properties: Optional[Dict[str, Any]],
        remote_idle_timeout_empty_frame_send_ratio: float,
        error_policy: Any,
        debug: bool,
        encoding: str,
        **kwargs: Any,
    ) -> Union[uamqp_Connection, pyamqp_Connection]:
        """
        Creates and returns the uamqp async Connection object.
        :keyword str endpoint: The endpoint, used by pyamqp.
        :keyword JWTTokenAuth auth: The auth, used by uamqp.
        :keyword str container_id: Required.
        :keyword int max_frame_size: Required.
        :keyword int channel_max: Required.
        :keyword float idle_timeout: Required.
        :keyword dict[str, Any] or None properties: Required.
        :keyword float remote_idle_timeout_empty_frame_send_ratio: Required.
        :keyword error_policy: Required.
        :keyword bool debug: Required.
        :keyword str encoding: Required.
        """

    @staticmethod
    @abstractmethod
    async def close_connection_async(connection: Union[uamqp_Connection, pyamqp_Connection]) -> None:
        """
        Closes existing connection.
        :param ~uamqp.async_ops.ConnectionAsync or ~pyamqp.aio.Connection connection: uamqp or pyamqp Connection.
        """

    @staticmethod
    @abstractmethod
    def get_connection_state(
        connection: Union[uamqp_Connection, pyamqp_Connection]
    ) -> Union["uamqp_ConnectionState", pyamqp_ConnectionState]:
        """
        Gets connection state.
        :param ~uamqp.async_ops.ConnectionAsync or ~pyamqp.aio.Connection connection: uamqp or pyamqp Connection.
        """

    @staticmethod
    @abstractmethod
    def create_send_client(
        *,
        config: Configuration,
        target: str,
        auth: Union["uamqp_JWTTokenAuth", "pyamqp_JWTTokenAuth"],
        idle_timeout: Optional[float],
        network_trace: bool,
        retry_policy: Any,
        keep_alive_interval: int,
        client_name: str,
        link_properties: Optional[Dict[str, Any]],
        properties: Optional[Dict[str, Any]],
        **kwargs: Any,
    ) -> Union[uamqp_SendClient, pyamqp_SendClient]:
        """
        Creates and returns the send client.
        :keyword ~azure.eventhub._configuration.Configuration config: The configuration.

        :keyword str target: Required. The target.
        :keyword JWTTokenAuth auth: Required.
        :keyword int idle_timeout: Required.
        :keyword bool network_trace: Required.
        :keyword retry_policy: Required.
        :keyword keep_alive_interval: Required.
        :keyword str client_name: Required.
        :keyword dict link_properties: Required.
        :keyword dict[str, Any] or None properties: Required.
        """

    @staticmethod
    @abstractmethod
    async def send_messages_async(
        producer: EventHubProducer, timeout_time: Optional[float], last_exception: Optional[Exception], logger: Logger
    ) -> None:
        """
        Handles sending of event data messages.
        :param ~azure.eventhub._producer.EventHubProducer producer: The producer with handler to send messages.
        :param int timeout_time: Timeout time.
        :param Exception last_exception: Exception to raise if message timed out. Only used by uamqp transport.
        :param logging.Logger logger: Logger.
        """

    @staticmethod
    @abstractmethod
    def set_message_partition_key(
        message: Union["uamqp_Message", "pyamqp_Message"], partition_key: Optional[Union[str, bytes]], **kwargs: Any
    ) -> Union["uamqp_Message", "pyamqp_Message"]:
        """Set the partition key as an annotation on a uamqp message.

        :param ~uamqp.Message or ~pyamqp.message.Message message: The message to update.
        :param str partition_key: The partition key value.
        :rtype: None
        """

    @staticmethod
    @abstractmethod
    def create_source(source: str, offset: int, selector: bytes) -> Union["uamqp_Source", "pyamqp_Source"]:
        """
        Creates and returns the Source.

        :param str source: Required.
        :param int offset: Required.
        :param bytes selector: Required.
        """

    @staticmethod
    @abstractmethod
    def create_receive_client(
        *,
        config: Configuration,
        source: Union["uamqp_Source", "pyamqp_Source"],
        auth: Union["uamqp_JWTTokenAuth", "pyamqp_JWTTokenAuth"],
        idle_timeout: Optional[float],
        network_trace: bool,
        retry_policy: Any,
        client_name: str,
        link_properties: Dict[bytes, Any],
        properties: Optional[Dict[str, Any]],
        link_credit: int,
        keep_alive_interval: int,
        desired_capabilities: Optional[List[bytes]],
        streaming_receive: bool,
        message_received_callback: Callable,
        timeout: int,
        **kwargs: Any,
    ) -> Union["uamqp_ReceiveClient", "pyamqp_ReceiveClient"]:
        """
        Creates and returns the receive client.
        :keyword ~azure.eventhub._configuration.Configuration config: The configuration.

        :keyword ~uamqp.Source or ~pyamqp.endpoints.Source source: Required. The source.
        :keyword JWTTokenAuth auth: Required.
        :keyword int idle_timeout: Required.
        :keyword network_trace: Required.
        :keyword retry_policy: Required.
        :keyword str client_name: Required.
        :keyword dict link_properties: Required.
        :keyword dict[str, Any] or None properties: Required.
        :keyword link_credit: Required. The prefetch.
        :keyword keep_alive_interval: Required. Missing in pyamqp.
        :keyword list[bytes] or None desired_capabilities: Required.
        :keyword streaming_receive: Required.
        :keyword message_received_callback: Required.
        :keyword timeout: Required.
        """

    @staticmethod
    @abstractmethod
    async def receive_messages_async(
        consumer: EventHubConsumer, batch: bool, max_batch_size: int, max_wait_time: Optional[int]
    ) -> None:
        """
        Receives messages, creates events, and returns them by calling the on received callback.
        :param ~azure.eventhub.aio._consumer_async.EventHubConsumer consumer: The EventHubConsumer.
        :param bool batch: If receive batch or single event.
        :param int max_batch_size: Max batch size.
        :param int max_wait_time: Max wait time.
        """

    @staticmethod
    @abstractmethod
    async def create_token_auth_async(
        auth_uri: str,
        get_token: Callable,
        token_type: bytes,
        config: Configuration,
        *,
        update_token: bool,
    ) -> Union["uamqp_JWTTokenAuth", "pyamqp_JWTTokenAuth"]:
        """
        Creates the JWTTokenAuth.
        :param str auth_uri: The auth uri to pass to JWTTokenAuth.
        :param callable get_token: The callback function used for getting and refreshing
         tokens. It should return a valid jwt token each time it is called.
        :param bytes token_type: Token type.
        :param ~azure.eventhub._configuration.Configuration config: EH config.

        :keyword bool update_token: Whether to update token. If not updating token,
         then pass 300 to refresh_window. Only used by uamqp.
        """

    @staticmethod
    @abstractmethod
    def create_mgmt_client(
        address: _Address, mgmt_auth: Union["uamqp_JWTTokenAuth", "pyamqp_JWTTokenAuth"], config: Configuration
    ) -> Union["uamqp_AMQPClient", "pyamqp_AMQPClient"]:
        """
        Creates and returns the mgmt AMQP client.
        :param _Address address: Required. The Address.
        :param JWTTokenAuth mgmt_auth: Auth for client.
        :param ~azure.eventhub._configuration.Configuration config: The configuration.
        """

    @staticmethod
    @abstractmethod
    async def get_updated_token_async(
        mgmt_auth: Union["uamqp_JWTTokenAuth", "pyamqp_JWTTokenAuth"]
    ) -> Union["uamqp_JWTTokenAuth", "pyamqp_JWTTokenAuth"]:
        """
        Return updated auth token.
        :param mgmt_auth: Auth.
        :type mgmt_auth: ~pyamqp.aio._authentication_async.JWTTokenAuthAsync
         or uamqp.authentication.JWTTokenAsync
        """

    @staticmethod
    @abstractmethod
    async def mgmt_client_request_async(
        mgmt_client: Union["uamqp_AMQPClient", "pyamqp_AMQPClient"],
        mgmt_msg: str,
        *,
        operation: bytes,
        operation_type: bytes,
        status_code_field: bytes,
        description_fields: bytes,
        **kwargs: Any,
    ) -> None:
        """
        Send mgmt request.
        :param ~uamqp.AMQPClientAsync or ~pyamqp.aio.AMQPClientAsync mgmt_client: Client to send request with.
        :param str mgmt_msg: Message.
        :keyword bytes operation: Operation.
        :keyword bytes operation_type: Op type.
        :keyword bytes status_code_field: mgmt status code.
        :keyword bytes description_fields: mgmt status desc.
        """

    @staticmethod
    @abstractmethod
    def get_error(status_code: int, description: str):
        """
        Gets error corresponding to status code.
        :param int status_code: Status code.
        :param str description: Description of error.
        """

    @staticmethod
    @abstractmethod
    def check_timeout_exception(base: ClientBase, exception: Exception):
        """
        Checks if timeout exception.
        :param ~azure.eventhub._client_base.ClientBase base: ClientBase.
        :param Exception exception: Exception to check.
        """



import time
import pytest
from azure.eventhub.aio._eventprocessor.in_memory_checkpoint_store import InMemoryCheckpointStore


TEST_NAMESPACE = "test_namespace"
TEST_EVENTHUB = "test_eventhub"
TEST_CONSUMER_GROUP = "test_consumer_group"
TEST_OWNER = "test_owner_id"


@pytest.mark.asyncio
async def test_claim_new_ownership():
    ownership = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "0",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP,
        "owner_id": "owner_0",
    }
    checkpoint_store = InMemoryCheckpointStore()
    claimed_ownership = await checkpoint_store.claim_ownership([ownership])
    assert len(claimed_ownership) == 1
    listed_ownership = await checkpoint_store.list_ownership(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP)

    assert listed_ownership[0] == ownership
    assert ownership["last_modified_time"] is not None
    assert ownership["etag"] is not None


@pytest.mark.asyncio
async def test_claim_existing_ownership():
    ownership = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "0",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP,
        "owner_id": "owner_0",
    }
    checkpoint_store = InMemoryCheckpointStore()
    await checkpoint_store.claim_ownership([ownership])
    listed_ownership = await checkpoint_store.list_ownership(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP)
    existing_ownership = listed_ownership[0]
    etag_existing = existing_ownership["etag"]
    last_modified_existing = existing_ownership["last_modified_time"]

    time.sleep(0.01)  # so time is changed
    copied_existing_ownership = existing_ownership.copy()
    copied_existing_ownership["owner_id"] = "owner_1"
    await checkpoint_store.claim_ownership([copied_existing_ownership])
    listed_ownership2 = await checkpoint_store.list_ownership(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP)

    assert len(listed_ownership2) == 1
    assert listed_ownership2[0] == copied_existing_ownership
    assert listed_ownership2[0]["owner_id"] == "owner_1"
    assert listed_ownership2[0]["etag"] != etag_existing
    assert listed_ownership2[0]["last_modified_time"] != last_modified_existing


@pytest.mark.asyncio
async def test_claim_two_ownerships():
    ownership_0 = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "0",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP,
        "owner_id": "owner_0",
    }
    ownership_1 = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "0",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP + "1",
        "owner_id": "owner_1",
    }
    checkpoint_store = InMemoryCheckpointStore()
    claimed_ownership = await checkpoint_store.claim_ownership([ownership_0, ownership_1])
    assert len(claimed_ownership) == 2
    listed_ownership0 = await checkpoint_store.list_ownership(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP)
    listed_ownership1 = await checkpoint_store.list_ownership(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP + "1")
    assert listed_ownership0[0]["consumer_group"] == TEST_CONSUMER_GROUP
    assert listed_ownership1[0]["consumer_group"] == TEST_CONSUMER_GROUP + "1"


@pytest.mark.asyncio
async def test_claim_two_partitions():
    ownership_0 = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "0",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP,
        "owner_id": "owner_0",
    }
    ownership_1 = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "1",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP,
        "owner_id": "owner_0",
    }
    checkpoint_store = InMemoryCheckpointStore()
    await checkpoint_store.claim_ownership([ownership_0, ownership_1])
    listed_ownership0 = await checkpoint_store.list_ownership(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP)
    assert set(x["partition_id"] for x in listed_ownership0) == {"0", "1"}


@pytest.mark.asyncio
async def test_update_checkpoint():
    checkpoint = {
        "fully_qualified_namespace": TEST_NAMESPACE,
        "partition_id": "0",
        "eventhub_name": TEST_EVENTHUB,
        "consumer_group": TEST_CONSUMER_GROUP,
        "offset": "0",
        "sequencenumber": 0,
    }
    checkpoint_store = InMemoryCheckpointStore()
    await checkpoint_store.update_checkpoint(checkpoint)
    listed_checkpoint = await checkpoint_store.list_checkpoints(TEST_NAMESPACE, TEST_EVENTHUB, TEST_CONSUMER_GROUP)

    assert listed_checkpoint[0] == checkpoint
    assert listed_checkpoint[0]["offset"] == "0"
    assert listed_checkpoint[0]["sequencenumber"] == 0



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import datetime
from typing import TYPE_CHECKING
from base64 import b64encode
from hashlib import sha256
from hmac import HMAC
from urllib.parse import urlencode, quote_plus
import time
from datetime import timezone
from .types import TYPE, VALUE, AMQPTypes
from ._encode import encode_payload
from .message import Properties

if TYPE_CHECKING:
    from .message import Message

TZ_UTC: timezone = timezone.utc
# Number of seconds between the Unix epoch (1/1/1970) and year 1 CE.
# This is the lowest value that can be represented by an AMQP timestamp.
CE_ZERO_SECONDS: int = -62_135_596_800


def utc_from_timestamp(timestamp: float) -> datetime.datetime:
    """
    :param float timestamp: Timestamp in seconds to be converted to datetime.
    :rtype: datetime.datetime
    :returns: A datetime object representing the timestamp in UTC.
    """
    # The AMQP timestamp is the number of seconds since the Unix epoch.
    # AMQP brokers represent the lowest value as -62_135_596_800 (the
    # number of seconds between the Unix epoch (1/1/1970) and year 1 CE) as
    # a sentinel for a time which is not set.
    if timestamp == CE_ZERO_SECONDS:
        return datetime.datetime.min.replace(tzinfo=TZ_UTC)
    return datetime.datetime.fromtimestamp(timestamp, tz=TZ_UTC)


def utc_now():
    return datetime.datetime.now(tz=TZ_UTC)


def encode(value, encoding="UTF-8"):
    return value.encode(encoding) if isinstance(value, str) else value


def generate_sas_token(audience, policy, key, expiry=None):
    """
    Generate a sas token according to the given audience, policy, key and expiry

    :param str audience:
    :param str policy:
    :param str key:
    :param int expiry: abs expiry time
    :return: A sas token
    :rtype: str
    """
    if not expiry:
        expiry = int(time.time()) + 3600  # Default to 1 hour.

    encoded_uri = quote_plus(audience)
    encoded_policy = quote_plus(policy).encode("utf-8")
    encoded_key = key.encode("utf-8")

    ttl = int(expiry)
    sign_key = "%s\n%d" % (encoded_uri, ttl)
    signature = b64encode(HMAC(encoded_key, sign_key.encode("utf-8"), sha256).digest())
    result = {"sr": audience, "sig": signature, "se": str(ttl)}
    if policy:
        result["skn"] = encoded_policy
    return "SharedAccessSignature " + urlencode(result)


def add_batch(batch, message):
    # Add a message to a batch
    output = bytearray()
    encode_payload(output, message)
    batch[5].append(output)

def set_message_properties(message, properties: list):
    if not message[3]:
        message[3] = Properties(*properties)

def set_message_annotations(message, annotations: dict):
    if not message[2]:
        message[2] = annotations

def encode_str(data, encoding="utf-8"):
    try:
        return data.encode(encoding)
    except AttributeError:
        return data


def normalized_data_body(data, **kwargs):
    # A helper method to normalize input into AMQP Data Body format
    encoding = kwargs.get("encoding", "utf-8")
    if isinstance(data, list):
        return [encode_str(item, encoding) for item in data]
    return [encode_str(data, encoding)]


def normalized_sequence_body(sequence):  # pylint:disable=inconsistent-return-statements
    # A helper method to normalize input into AMQP Sequence Body format
    if isinstance(sequence, list) and all((isinstance(b, list) for b in sequence)):
        return sequence
    if isinstance(sequence, list):
        return [sequence]


def get_message_encoded_size(message):
    output = bytearray()
    encode_payload(output, message)
    return len(output)


def amqp_long_value(value):
    # A helper method to wrap a Python int as AMQP long
    # TODO: wrapping one line in a function is expensive, find if there's a better way to do it
    return {TYPE: AMQPTypes.long, VALUE: value}


def amqp_uint_value(value):
    # A helper method to wrap a Python int as AMQP uint
    return {TYPE: AMQPTypes.uint, VALUE: value}


def amqp_string_value(value):
    return {TYPE: AMQPTypes.string, VALUE: value}


def amqp_symbol_value(value):
    return {TYPE: AMQPTypes.symbol, VALUE: value}


def amqp_array_value(value):
    return {TYPE: AMQPTypes.array, VALUE: value}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from __future__ import annotations
import uuid
import logging
import time
from typing import Union, Optional

from .constants import ConnectionState, SessionState, SessionTransferState, Role
from .sender import SenderLink
from .receiver import ReceiverLink
from .management_link import ManagementLink
from .performatives import (
    BeginFrame,
    EndFrame,
    FlowFrame,
    TransferFrame,
    DispositionFrame,
)
from .error import AMQPError, ErrorCondition
from ._encode import encode_frame

_LOGGER = logging.getLogger(__name__)


class Session(object):  # pylint: disable=too-many-instance-attributes
    """
    :param int remote_channel: The remote channel for this Session.
    :param int next_outgoing_id: The transfer-id of the first transfer id the sender will send.
    :param int incoming_window: The initial incoming-window of the sender.
    :param int outgoing_window: The initial outgoing-window of the sender.
    :param int handle_max: The maximum handle value that may be used on the Session.
    :param list(str) offered_capabilities: The extension capabilities the sender supports.
    :param list(str) desired_capabilities: The extension capabilities the sender may use if the receiver supports
    :param dict properties: Session properties.
    """

    def __init__(self, connection, channel, **kwargs):
        self.name = kwargs.pop("name", None) or str(uuid.uuid4())
        self.state = SessionState.UNMAPPED
        self.handle_max = kwargs.get("handle_max", 4294967295)
        self.properties = kwargs.pop("properties", None)
        self.remote_properties = None
        self.channel = channel
        self.remote_channel = None
        self.next_outgoing_id = kwargs.pop("next_outgoing_id", 0)
        self.next_incoming_id = None
        self.incoming_window = kwargs.pop("incoming_window", 1)
        self.outgoing_window = kwargs.pop("outgoing_window", 1)
        self.target_incoming_window = self.incoming_window
        self.remote_incoming_window = 0
        self.remote_outgoing_window = 0
        self.offered_capabilities = None
        self.desired_capabilities = kwargs.pop("desired_capabilities", None)

        self.allow_pipelined_open = kwargs.pop("allow_pipelined_open", True)
        self.idle_wait_time = kwargs.get("idle_wait_time", 0.1)
        self.network_trace = kwargs["network_trace"]
        self.network_trace_params = kwargs["network_trace_params"]
        self.network_trace_params["amqpSession"] = self.name

        self.links = {}
        self._connection = connection
        self._output_handles = {}
        self._input_handles = {}

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, *args):
        self.end()

    @classmethod
    def from_incoming_frame(cls, connection, channel):
        # TODO: check session_create_from_endpoint in C lib
        new_session = cls(connection, channel)
        return new_session

    def _set_state(self, new_state: SessionState) -> None:
        """Update the session state.
        :param ~pyamqp.constants.SessionState new_state: The new state to set.
        """
        if new_state is None:
            return
        previous_state = self.state
        self.state = new_state
        _LOGGER.info(
            "Session state changed: %r -> %r",
            previous_state,
            new_state,
            extra=self.network_trace_params,
        )
        for link in self.links.values():
            link._on_session_state_change()  # pylint: disable=protected-access

    def _on_connection_state_change(self):
        if self._connection.state in [ConnectionState.CLOSE_RCVD, ConnectionState.END]:
            if self.state not in [SessionState.DISCARDING, SessionState.UNMAPPED]:
                self._set_state(SessionState.DISCARDING)

    def _get_next_output_handle(self) -> int:
        """Get the next available outgoing handle number within the max handle limit.

        :raises ValueError: If maximum handle has been reached.
        :returns: The next available outgoing handle number.
        :rtype: int
        """
        if len(self._output_handles) >= self.handle_max:
            raise ValueError("Maximum number of handles ({}) has been reached.".format(self.handle_max))
        next_handle = next(i for i in range(1, self.handle_max) if i not in self._output_handles)
        return next_handle

    def _outgoing_begin(self):
        begin_frame = BeginFrame(
            remote_channel=self.remote_channel if self.state == SessionState.BEGIN_RCVD else None,
            next_outgoing_id=self.next_outgoing_id,
            outgoing_window=self.outgoing_window,
            incoming_window=self.incoming_window,
            handle_max=self.handle_max,
            offered_capabilities=self.offered_capabilities if self.state == SessionState.BEGIN_RCVD else None,
            desired_capabilities=self.desired_capabilities if self.state == SessionState.UNMAPPED else None,
            properties=self.properties,
        )
        if self.network_trace:
            _LOGGER.debug("-> %r", begin_frame, extra=self.network_trace_params)
        self._connection._process_outgoing_frame(self.channel, begin_frame)  # pylint: disable=protected-access

    def _incoming_begin(self, frame):
        if self.network_trace:
            _LOGGER.debug("<- %r", BeginFrame(*frame), extra=self.network_trace_params)
        self.handle_max = frame[4]  # handle_max
        self.next_incoming_id = frame[1]  # next_outgoing_id
        self.remote_incoming_window = frame[2]  # incoming_window
        self.remote_outgoing_window = frame[3]  # outgoing_window
        self.remote_properties = frame[7]  # incoming map of properties about the session
        if self.state == SessionState.BEGIN_SENT:
            self.remote_channel = frame[0]  # remote_channel
            self._set_state(SessionState.MAPPED)
        elif self.state == SessionState.UNMAPPED:
            self._set_state(SessionState.BEGIN_RCVD)
            self._outgoing_begin()
            self._set_state(SessionState.MAPPED)

    def _outgoing_end(self, error=None):
        end_frame = EndFrame(error=error)
        if self.network_trace:
            _LOGGER.debug("-> %r", end_frame, extra=self.network_trace_params)
        self._connection._process_outgoing_frame(self.channel, end_frame)  # pylint: disable=protected-access

    def _incoming_end(self, frame):
        if self.network_trace:
            _LOGGER.debug("<- %r", EndFrame(*frame), extra=self.network_trace_params)
        if self.state not in [
            SessionState.END_RCVD,
            SessionState.END_SENT,
            SessionState.DISCARDING,
        ]:
            self._set_state(SessionState.END_RCVD)
            for _, link in self.links.items():
                link.detach()
            # TODO: handling error
            self._outgoing_end()
        self._set_state(SessionState.UNMAPPED)

    def _outgoing_attach(self, frame):
        self._connection._process_outgoing_frame(self.channel, frame)  # pylint: disable=protected-access

    def _incoming_attach(self, frame):
        try:
            self._input_handles[frame[1]] = self.links[frame[0].decode("utf-8")]  # name and handle
            self._input_handles[frame[1]]._incoming_attach(frame)  # pylint: disable=protected-access
        except KeyError:
            try:
                outgoing_handle = self._get_next_output_handle()
            except ValueError:
                _LOGGER.error(
                    "Unable to attach new link - cannot allocate more handles.", extra=self.network_trace_params
                )
                # detach the link that would have been set.
                self.links[frame[0].decode("utf-8")].detach(
                    error=AMQPError(
                        condition=ErrorCondition.LinkDetachForced,
                        description="""Cannot allocate more handles, """
                        """the max number of handles is {}. Detaching link""".format(self.handle_max),
                        info=None,
                    )
                )
                return
            if frame[2] == Role.Sender:  # role
                new_link = ReceiverLink.from_incoming_frame(self, outgoing_handle, frame)
            else:
                new_link = SenderLink.from_incoming_frame(self, outgoing_handle, frame)
            new_link._incoming_attach(frame)  # pylint: disable=protected-access
            self.links[frame[0]] = new_link
            self._output_handles[outgoing_handle] = new_link
            self._input_handles[frame[1]] = new_link
        except ValueError as e:
            # Reject Link
            _LOGGER.debug("Unable to attach new link: %r", e, extra=self.network_trace_params)
            self._input_handles[frame[1]].detach()

    def _outgoing_flow(self, frame=None):
        link_flow = frame or {}
        link_flow.update(
            {
                "next_incoming_id": self.next_incoming_id,
                "incoming_window": self.incoming_window,
                "next_outgoing_id": self.next_outgoing_id,
                "outgoing_window": self.outgoing_window,
            }
        )
        flow_frame = FlowFrame(**link_flow)
        if self.network_trace:
            _LOGGER.debug("-> %r", flow_frame, extra=self.network_trace_params)
        self._connection._process_outgoing_frame(self.channel, flow_frame)  # pylint: disable=protected-access

    def _incoming_flow(self, frame):
        if self.network_trace:
            _LOGGER.debug("<- %r", FlowFrame(*frame), extra=self.network_trace_params)
        self.next_incoming_id = frame[2]  # next_outgoing_id
        remote_incoming_id = frame[0] or self.next_outgoing_id  #  next_incoming_id  TODO "initial-outgoing-id"
        self.remote_incoming_window = remote_incoming_id + frame[1] - self.next_outgoing_id  # incoming_window
        self.remote_outgoing_window = frame[3]  # outgoing_window
        if frame[4] is not None:  # handle
            self._input_handles[frame[4]]._incoming_flow(frame)  # pylint: disable=protected-access
        else:
            for link in self._output_handles.values():
                if self.remote_incoming_window > 0 and not link._is_closed:  # pylint: disable=protected-access
                    link._incoming_flow(frame)  # pylint: disable=protected-access

    def _outgoing_transfer(self, delivery, network_trace_params):
        if self.state != SessionState.MAPPED:
            delivery.transfer_state = SessionTransferState.ERROR
        if self.remote_incoming_window <= 0:
            delivery.transfer_state = SessionTransferState.BUSY
        else:
            payload = delivery.frame["payload"]
            payload_size = len(payload)

            delivery.frame["delivery_id"] = self.next_outgoing_id
            # calculate the transfer frame encoding size excluding the payload
            delivery.frame["payload"] = b""
            # TODO: encoding a frame would be expensive, we might want to improve depending on the perf test results
            encoded_frame = encode_frame(TransferFrame(**delivery.frame))[1]
            transfer_overhead_size = len(encoded_frame)

            # available size for payload per frame is calculated as following:
            # remote max frame size - transfer overhead (calculated) - header (8 bytes)
            available_frame_size = (
                self._connection._remote_max_frame_size - transfer_overhead_size - 8  # pylint: disable=protected-access
            )

            start_idx = 0
            remaining_payload_cnt = payload_size
            # encode n-1 frames if payload_size > available_frame_size
            while remaining_payload_cnt > available_frame_size:
                tmp_delivery_frame = {
                    "handle": delivery.frame["handle"],
                    "delivery_tag": delivery.frame["delivery_tag"],
                    "message_format": delivery.frame["message_format"],
                    "settled": delivery.frame["settled"],
                    "more": True,
                    "rcv_settle_mode": delivery.frame["rcv_settle_mode"],
                    "state": delivery.frame["state"],
                    "resume": delivery.frame["resume"],
                    "aborted": delivery.frame["aborted"],
                    "batchable": delivery.frame["batchable"],
                    "delivery_id": self.next_outgoing_id,
                }
                if network_trace_params:
                    # We determine the logging for the outgoing Transfer frames based on the source
                    # Link configuration rather than the Session, because it's only at the Session
                    # level that we can determine how many outgoing frames are needed and their
                    # delivery IDs.
                    # TODO: Obscuring the payload for now to investigate the potential for leaks.
                    _LOGGER.debug(
                        "-> %r", TransferFrame(payload=b"***", **tmp_delivery_frame), extra=network_trace_params
                    )
                self._connection._process_outgoing_frame(  # pylint: disable=protected-access
                    self.channel,
                    TransferFrame(payload=payload[start_idx : start_idx + available_frame_size], **tmp_delivery_frame),
                )
                start_idx += available_frame_size
                remaining_payload_cnt -= available_frame_size

            # encode the last frame
            tmp_delivery_frame = {
                "handle": delivery.frame["handle"],
                "delivery_tag": delivery.frame["delivery_tag"],
                "message_format": delivery.frame["message_format"],
                "settled": delivery.frame["settled"],
                "more": False,
                "rcv_settle_mode": delivery.frame["rcv_settle_mode"],
                "state": delivery.frame["state"],
                "resume": delivery.frame["resume"],
                "aborted": delivery.frame["aborted"],
                "batchable": delivery.frame["batchable"],
                "delivery_id": self.next_outgoing_id,
            }
            if network_trace_params:
                # We determine the logging for the outgoing Transfer frames based on the source
                # Link configuration rather than the Session, because it's only at the Session
                # level that we can determine how many outgoing frames are needed and their
                # delivery IDs.
                # TODO: Obscuring the payload for now to investigate the potential for leaks.
                _LOGGER.debug("-> %r", TransferFrame(payload=b"***", **tmp_delivery_frame), extra=network_trace_params)
            self._connection._process_outgoing_frame(  # pylint: disable=protected-access
                self.channel, TransferFrame(payload=payload[start_idx:], **tmp_delivery_frame)
            )
            self.next_outgoing_id += 1
            self.remote_incoming_window -= 1
            self.outgoing_window -= 1
            # TODO: We should probably handle an error at the connection and update state accordingly
            delivery.transfer_state = SessionTransferState.OKAY

    def _incoming_transfer(self, frame):
        # TODO: should this be only if more=False?
        self.next_incoming_id += 1
        self.remote_outgoing_window -= 1
        self.incoming_window -= 1
        try:
            self._input_handles[frame[0]]._incoming_transfer(frame)  # pylint: disable=protected-access
        except KeyError:
            _LOGGER.error(
                "Received Transfer frame on unattached link. Ending session.", extra=self.network_trace_params
            )
            self._set_state(SessionState.DISCARDING)
            self.end(
                error=AMQPError(
                    condition=ErrorCondition.SessionUnattachedHandle,
                    description="""Invalid handle reference in received frame: """
                    """Handle is not currently associated with an attached link""",
                )
            )
            return
        if self.incoming_window == 0:
            self.incoming_window = self.target_incoming_window
            self._outgoing_flow()

    def _outgoing_disposition(self, frame):
        self._connection._process_outgoing_frame(self.channel, frame)  # pylint: disable=protected-access

    def _incoming_disposition(self, frame):
        if self.network_trace:
            _LOGGER.debug("<- %r", DispositionFrame(*frame), extra=self.network_trace_params)
        for link in self._input_handles.values():
            link._incoming_disposition(frame)  # pylint: disable=protected-access

    def _outgoing_detach(self, frame):
        self._connection._process_outgoing_frame(self.channel, frame)  # pylint: disable=protected-access

    def _incoming_detach(self, frame):
        try:
            link = self._input_handles[frame[0]]  # handle
            link._incoming_detach(frame)  # pylint: disable=protected-access
            # if link._is_closed:  TODO
            #     self.links.pop(link.name, None)
            #     self._input_handles.pop(link.remote_handle, None)
            #     self._output_handles.pop(link.handle, None)
        except KeyError:
            self._set_state(SessionState.DISCARDING)
            self._connection.close(
                error=AMQPError(
                    condition=ErrorCondition.SessionUnattachedHandle,
                    description="""Invalid handle reference in received frame: """
                    """Handle is not currently associated with an attached link""",
                )
            )

    def _wait_for_response(self, wait: Union[bool, float], end_state: SessionState) -> None:
        if wait is True:
            self._connection.listen(wait=False)
            while self.state != end_state:
                time.sleep(self.idle_wait_time)
                self._connection.listen(wait=False)
        elif wait:
            self._connection.listen(wait=False)
            timeout = time.time() + wait
            while self.state != end_state:
                if time.time() >= timeout:
                    break
                time.sleep(self.idle_wait_time)
                self._connection.listen(wait=False)

    def begin(self, wait=False):
        self._outgoing_begin()
        self._set_state(SessionState.BEGIN_SENT)
        if wait:
            self._wait_for_response(wait, SessionState.BEGIN_SENT)
        elif not self.allow_pipelined_open:
            raise ValueError("Connection has been configured to not allow piplined-open. Please set 'wait' parameter.")

    def end(self, error: Optional[AMQPError] = None, wait: bool = False) -> None:
        try:
            if self.state not in [SessionState.UNMAPPED, SessionState.DISCARDING]:
                self._outgoing_end(error=error)
            for _, link in self.links.items():
                link.detach()
            new_state = SessionState.DISCARDING if error else SessionState.END_SENT
            self._set_state(new_state)
            self._wait_for_response(wait, SessionState.UNMAPPED)
        except Exception as exc:  # pylint: disable=broad-except
            _LOGGER.info("An error occurred when ending the session: %r", exc, extra=self.network_trace_params)
            self._set_state(SessionState.UNMAPPED)

    def create_receiver_link(self, source_address, **kwargs):
        assigned_handle = self._get_next_output_handle()
        link = ReceiverLink(
            self,
            handle=assigned_handle,
            source_address=source_address,
            network_trace=kwargs.pop("network_trace", self.network_trace),
            network_trace_params=dict(self.network_trace_params),
            **kwargs,
        )
        self.links[link.name] = link
        self._output_handles[assigned_handle] = link
        return link

    def create_sender_link(self, target_address, **kwargs):
        assigned_handle = self._get_next_output_handle()
        link = SenderLink(
            self,
            handle=assigned_handle,
            target_address=target_address,
            network_trace=kwargs.pop("network_trace", self.network_trace),
            network_trace_params=dict(self.network_trace_params),
            **kwargs,
        )
        self._output_handles[assigned_handle] = link
        self.links[link.name] = link
        return link

    def create_request_response_link_pair(self, endpoint, **kwargs):
        return ManagementLink(
            self,
            endpoint,
            network_trace=kwargs.pop("network_trace", self.network_trace),
            network_trace_params=dict(self.network_trace_params),
            **kwargs,
        )



# coding=utf-8
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import asyncio
from collections import defaultdict
import pytest
from azure.eventhub.aio._buffered_producer._partition_resolver_async import PartitionResolver


class TestPartitionResolver:

    @pytest.mark.asyncio
    @pytest.mark.parametrize("partition_cnt", [1, 2, 16, 32, 256])
    async def test_basic_round_robin(self, partition_cnt):
        partitions = [str(i) for i in range(partition_cnt)]
        pr = PartitionResolver(partitions)
        for i in range(2 * partition_cnt):
            expected = str(i % partition_cnt)
            real = await pr.get_next_partition_id()
            assert expected == real

    @pytest.mark.asyncio
    @pytest.mark.parametrize("partition_cnt", [1, 2, 16, 32, 256])
    async def test_concurrent_round_robin_fairly(self, partition_cnt):
        partitions = [str(i) for i in range(partition_cnt)]
        pr = PartitionResolver(partitions)
        dic = defaultdict(int)
        lock = asyncio.Lock()

        async def gen_pid():
            pid = await pr.get_next_partition_id()
            async with lock:
                dic[pid] += 1

        futures = [asyncio.ensure_future(gen_pid()) for _ in range(5 * partition_cnt)]

        for future in futures:
            await future

        assert len(dic) == partition_cnt
        for i in range(partition_cnt):
            assert dic[str(i)] == 5



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import pytest
import asyncio
import logging

from azure.eventhub._pyamqp.aio import _authentication_async
from azure.eventhub._pyamqp.aio import ReceiveClientAsync, SendClientAsync
from azure.eventhub._pyamqp.constants import TransportType
from azure.eventhub._pyamqp.message import Message


async def send_message(live_eventhub):
    uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
    sas_auth = _authentication_async.SASTokenAuthAsync(
        uri=uri, audience=uri, username=live_eventhub["key_name"], password=live_eventhub["access_key"]
    )

    target = "amqps://{}/{}/Partitions/{}".format(
        live_eventhub["hostname"], live_eventhub["event_hub"], live_eventhub["partition"]
    )

    message = Message(value="Single Message")

    async with SendClientAsync(
        live_eventhub["hostname"], target, auth=sas_auth, debug=True, transport_type=TransportType.Amqp
    ) as send_client:
        await send_client.send_message_async(message)


@pytest.mark.asyncio
async def test_event_hubs_client_amqp_async(live_eventhub):
    uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
    sas_auth = _authentication_async.SASTokenAuthAsync(
        uri=uri, audience=uri, username=live_eventhub["key_name"], password=live_eventhub["access_key"]
    )

    source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
        live_eventhub["hostname"],
        live_eventhub["event_hub"],
        live_eventhub["consumer_group"],
        live_eventhub["partition"],
    )

    await send_message(live_eventhub=live_eventhub)

    async with ReceiveClientAsync(
        live_eventhub["hostname"],
        source,
        auth=sas_auth,
        debug=False,
        timeout=500,
        prefetch=1,
        transport_type=TransportType.Amqp,
    ) as receive_client:
        messages = await receive_client.receive_message_batch_async(max_batch_size=1)
        assert len(messages) > 0



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from enum import Enum


TYPE = "TYPE"
VALUE = "VALUE"


class AMQPTypes:
    null = "NULL"
    boolean = "BOOL"
    ubyte = "UBYTE"
    byte = "BYTE"
    ushort = "USHORT"
    short = "SHORT"
    uint = "UINT"
    int = "INT"
    ulong = "ULONG"
    long = "LONG"
    float = "FLOAT"
    double = "DOUBLE"
    timestamp = "TIMESTAMP"
    uuid = "UUID"
    binary = "BINARY"
    string = "STRING"
    symbol = "SYMBOL"
    list = "LIST"
    map = "MAP"
    array = "ARRAY"
    described = "DESCRIBED"
    decimal128 = "DECIMAL128"


class FieldDefinition(Enum):
    fields = "fields"
    annotations = "annotations"
    message_id = "message-id"
    app_properties = "application-properties"
    node_properties = "node-properties"
    filter_set = "filter-set"


class ObjDefinition(Enum):
    source = "source"
    target = "target"
    delivery_state = "delivery-state"
    error = "error"


class ConstructorBytes(object):
    null = b"\x40"
    bool = b"\x56"
    bool_true = b"\x41"
    bool_false = b"\x42"
    ubyte = b"\x50"
    byte = b"\x51"
    ushort = b"\x60"
    short = b"\x61"
    uint_0 = b"\x43"
    uint_small = b"\x52"
    int_small = b"\x54"
    uint_large = b"\x70"
    int_large = b"\x71"
    ulong_0 = b"\x44"
    ulong_small = b"\x53"
    long_small = b"\x55"
    ulong_large = b"\x80"
    long_large = b"\x81"
    float = b"\x72"
    double = b"\x82"
    timestamp = b"\x83"
    uuid = b"\x98"
    binary_small = b"\xA0"
    binary_large = b"\xB0"
    string_small = b"\xA1"
    string_large = b"\xB1"
    symbol_small = b"\xA3"
    symbol_large = b"\xB3"
    list_0 = b"\x45"
    list_small = b"\xC0"
    list_large = b"\xD0"
    map_small = b"\xC1"
    map_large = b"\xD1"
    array_small = b"\xE0"
    array_large = b"\xF0"
    descriptor = b"\x00"
    decimal128 = b"\x94"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import annotations
import logging
from typing import Dict, Optional, Any, TYPE_CHECKING

from .._utils import get_last_enqueued_event_properties
from .checkpoint_store import CheckpointStore

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .._common import EventData


class PartitionContext:
    """Contains partition related context information.

    A `PartitionContext` instance will be passed to the event, error and initialization callbacks defined
    when calling `EventHubConsumerClient.receive()`.
    Users can call `update_checkpoint()` of this class to persist checkpoint data.
    """

    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        consumer_group: str,
        partition_id: str,
        checkpoint_store: Optional[CheckpointStore] = None,
    ) -> None:
        self.fully_qualified_namespace = fully_qualified_namespace
        self.partition_id = partition_id
        self.eventhub_name = eventhub_name
        self.consumer_group = consumer_group
        self._checkpoint_store = checkpoint_store
        self._last_received_event: Optional[EventData] = None

    @property
    def last_enqueued_event_properties(self) -> Optional[Dict[str, Any]]:
        """The latest enqueued event information.

        This property will be updated each time an event is received if the receiver is created
        with `track_last_enqueued_event_properties` set to `True`.
        The properties dict includes following information of the last enqueued event:

            - `sequence_number` (int)
            - `offset` (str)
            - `enqueued_time` (UTC datetime.datetime)
            - `retrieval_time` (UTC datetime.datetime)

        :rtype: Dict[str, Any] or None
        """
        if self._last_received_event:
            return get_last_enqueued_event_properties(self._last_received_event)
        return None

    def update_checkpoint(self, event: Optional[EventData] = None, **kwargs: Any) -> None:
        """Updates the receive checkpoint to the given events offset.

        :param ~azure.eventhub.EventData event: The EventData instance which contains the offset and
         sequence number information used for checkpoint.
        :rtype: None
        """
        if self._checkpoint_store:
            checkpoint_event = event or self._last_received_event
            if checkpoint_event:
                checkpoint = {
                    "fully_qualified_namespace": self.fully_qualified_namespace,
                    "eventhub_name": self.eventhub_name,
                    "consumer_group": self.consumer_group,
                    "partition_id": self.partition_id,
                    "offset": checkpoint_event.offset,
                    "sequence_number": checkpoint_event.sequence_number,
                }
                try:
                    self._checkpoint_store.update_checkpoint(checkpoint, **kwargs)
                except TypeError as e:
                    if "update_checkpoint() got an unexpected keyword argument" in str(e):
                        _LOGGER.info(
                            "The provided checkpointstore method 'update_checkpoint' does not accept keyword arguments,"
                            " so keyword arguments will be ignored. Please update method signature to support kwargs."
                        )
                        self._checkpoint_store.update_checkpoint(checkpoint)
                    else:
                        raise e
        else:
            _LOGGER.warning(
                "namespace %r, eventhub %r, consumer_group %r, partition_id %r "
                "update_checkpoint is called without checkpoint store. No checkpoint is updated.",
                self.fully_qualified_namespace,
                self.eventhub_name,
                self.consumer_group,
                self.partition_id,
            )



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to show streaming sending events with different options to an Event Hub asynchronously.
"""

import time
import asyncio
import os

from azure.eventhub.aio import EventHubProducerClient
from azure.eventhub import EventData
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


async def run():

    producer = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
    )

    to_send_message_cnt = 500
    bytes_per_message = 256

    async with producer:
        event_data_batch = await producer.create_batch()
        for i in range(to_send_message_cnt):
            event_data = EventData("D" * bytes_per_message)
            try:
                event_data_batch.add(event_data)
            except ValueError:
                await producer.send_batch(event_data_batch)
                event_data_batch = await producer.create_batch()
                event_data_batch.add(event_data)
        if len(event_data_batch) > 0:
            await producer.send_batch(event_data_batch)


start_time = time.time()
asyncio.run(run())
print("Send messages in {} seconds.".format(time.time() - start_time))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._client_base_async import EventHubSharedKeyCredential
from ._consumer_client_async import EventHubConsumerClient
from ._producer_client_async import EventHubProducerClient
from ._eventprocessor.checkpoint_store import CheckpointStore
from ._eventprocessor.partition_context import PartitionContext

__all__ = [
    "EventHubSharedKeyCredential",
    "EventHubConsumerClient",
    "EventHubProducerClient",
    "CheckpointStore",
    "PartitionContext",
]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations
import random
import uuid
import logging
import time
import threading
from typing import (
    Dict,
    Callable,
    List,
    Any,
    Mapping,
    Union,
    TYPE_CHECKING,
    Optional,
    Iterable,
    cast,
)
from functools import partial

from .._tracing import (
    process_context_manager,
    receive_context_manager,
    get_span_links_from_received_events,
    is_tracing_enabled,
)
from .partition_context import PartitionContext
from .in_memory_checkpoint_store import InMemoryCheckpointStore
from .ownership_manager import OwnershipManager
from .common import CloseReason, LoadBalancingStrategy
from ._eventprocessor_mixin import EventProcessorMixin

if TYPE_CHECKING:
    from datetime import datetime
    from .checkpoint_store import CheckpointStore
    from .._common import EventData
    from .._consumer import EventHubConsumer
    from .._consumer_client import EventHubConsumerClient


_LOGGER = logging.getLogger(__name__)


class EventProcessor(EventProcessorMixin):  # pylint:disable=too-many-instance-attributes
    """
    An EventProcessor constantly receives events from one or multiple partitions of the Event Hub
    in the context of a given consumer group.

    """

    def __init__(
        self,
        eventhub_client: EventHubConsumerClient,
        consumer_group: str,
        on_event: Union[
            Callable[["PartitionContext", Optional["EventData"]], None],
            Callable[["PartitionContext", List["EventData"]], None],
        ],
        *,
        batch: bool = False,
        max_batch_size: int = 300,
        max_wait_time: Optional[float] = None,
        partition_id: Optional[str] = None,
        on_error: Optional[Callable[[PartitionContext, Exception], None]] = None,
        on_partition_initialize: Optional[Callable[[PartitionContext], None]] = None,
        on_partition_close: Optional[Callable[[PartitionContext, CloseReason], None]] = None,
        checkpoint_store: Optional[CheckpointStore] = None,
        initial_event_position: Union[int, str, datetime, Mapping[str, Union[int, str, datetime]]] = "@latest",
        initial_event_position_inclusive: Union[bool, Mapping[str, bool]] = False,
        load_balancing_interval: float = 30.0,
        load_balancing_strategy: LoadBalancingStrategy = LoadBalancingStrategy.GREEDY,
        partition_ownership_expiration_interval: Optional[float] = None,
        owner_level: Optional[int] = None,
        prefetch: Optional[int] = None,
        track_last_enqueued_event_properties: bool = False,
    ) -> None:
        self._consumer_group = consumer_group
        self._eventhub_client = eventhub_client
        self._namespace = eventhub_client._address.hostname
        self._eventhub_name = eventhub_client.eventhub_name
        self._event_handler = on_event
        self._batch = batch
        self._max_batch_size = max_batch_size
        self._max_wait_time = max_wait_time
        self._partition_id = partition_id
        self._error_handler: Optional[Callable[[PartitionContext, Exception], None]] = on_error
        self._partition_initialize_handler: Optional[Callable[[PartitionContext], None]] = on_partition_initialize
        self._partition_close_handler: Optional[Callable[[PartitionContext, CloseReason], None]] = on_partition_close
        self._checkpoint_store = checkpoint_store or InMemoryCheckpointStore()
        self._initial_event_position: Union[int, str, datetime, Mapping[str, Union[int, str, datetime]]] = (
            initial_event_position
        )
        self._initial_event_position_inclusive: Union[bool, Mapping[str, bool]] = initial_event_position_inclusive

        self._load_balancing_interval: float = load_balancing_interval
        self._load_balancing_strategy = load_balancing_strategy
        self._ownership_timeout = (
            partition_ownership_expiration_interval
            if partition_ownership_expiration_interval is not None
            else self._load_balancing_interval * 6
        )

        self._partition_contexts: Dict[str, PartitionContext] = {}

        # Receive parameters
        self._owner_level: Optional[int] = owner_level
        if checkpoint_store and self._owner_level is None:
            self._owner_level = 0
        self._prefetch: Optional[int] = prefetch
        self._track_last_enqueued_event_properties = track_last_enqueued_event_properties
        self._id = str(uuid.uuid4())
        self._running = False
        self._lock = threading.RLock()

        self._consumers: Dict[str, EventHubConsumer] = {}
        self._ownership_manager = OwnershipManager(
            self._eventhub_client,
            self._consumer_group,
            self._id,
            self._checkpoint_store,
            self._ownership_timeout,
            self._load_balancing_strategy,
            self._partition_id,
        )

        self._last_received_time = time.time_ns()

    def __repr__(self) -> str:
        return "EventProcessor: id {}".format(self._id)

    def _process_error(self, partition_context, err):
        if self._error_handler:
            try:
                self._error_handler(partition_context, err)
            except Exception as err_again:  # pylint:disable=broad-except
                _LOGGER.warning(
                    "EventProcessor instance %r of eventhub %r partition %r consumer group %r. "
                    "An error occurred while running on_error. The exception is %r.",
                    self._id,
                    partition_context.eventhub_name,
                    partition_context.partition_id,
                    partition_context.consumer_group,
                    err_again,
                )

    def _cancel_tasks_for_partitions(self, to_cancel_partitions: Iterable[str]) -> None:
        with self._lock:
            _LOGGER.debug(
                "EventProcessor %r tries to cancel partitions %r",
                self._id,
                to_cancel_partitions,
            )
            for partition_id in to_cancel_partitions:
                if partition_id in self._consumers:
                    self._consumers[partition_id].stop = True
                    _LOGGER.info(
                        "EventProcessor %r has cancelled partition %r",
                        self._id,
                        partition_id,
                    )

    def _initialize_partition_consumer(self, partition_id):
        if self._partition_initialize_handler:
            try:
                self._partition_initialize_handler(self._partition_contexts[partition_id])
            except Exception as err:  # pylint:disable=broad-except
                _LOGGER.warning(
                    "EventProcessor instance %r of eventhub %r partition %r consumer group %r. "
                    "An error occurred while running on_partition_initialize. The exception is %r.",
                    self._id,
                    self._partition_contexts[partition_id].eventhub_name,
                    self._partition_contexts[partition_id].partition_id,
                    self._partition_contexts[partition_id].consumer_group,
                    err,
                )
                self._process_error(self._partition_contexts[partition_id], err)
        _LOGGER.info("EventProcessor %r has claimed partition %r", self._id, partition_id)

    def _create_tasks_for_claimed_ownership(
        self, claimed_partitions: Iterable[str], checkpoints: Optional[Dict[str, Dict[str, Any]]] = None
    ) -> None:
        with self._lock:
            _LOGGER.debug(
                "EventProcessor %r tries to claim partition %r",
                self._id,
                claimed_partitions,
            )
            for partition_id in claimed_partitions:
                if partition_id not in self._consumers:
                    if partition_id in self._partition_contexts:
                        partition_context = self._partition_contexts[partition_id]
                        partition_context._last_received_event = None  # pylint:disable=protected-access
                    else:
                        partition_context = PartitionContext(
                            self._namespace,
                            self._eventhub_name,
                            self._consumer_group,
                            partition_id,
                            self._checkpoint_store,
                        )
                        self._partition_contexts[partition_id] = partition_context

                    checkpoint = checkpoints.get(partition_id) if checkpoints else None
                    (
                        initial_event_position,
                        event_postition_inclusive,
                    ) = self.get_init_event_position(partition_id, checkpoint)
                    event_received_callback = partial(self._on_event_received, partition_context)
                    self._consumers[partition_id] = cast(
                        "EventHubConsumer",
                        self.create_consumer(
                            partition_id,
                            initial_event_position,
                            event_postition_inclusive,
                            event_received_callback,
                        ),
                    )
                    self._initialize_partition_consumer(partition_id)

    def _on_event_received(
        self, partition_context: PartitionContext, event: Union[Optional[EventData], List[EventData]]
    ) -> None:
        if event:
            try:
                partition_context._last_received_event = event[-1]  # type: ignore  #pylint:disable=protected-access
            except TypeError:
                partition_context._last_received_event = event  # type: ignore  #pylint:disable=protected-access

            links = []
            is_batch = False
            if is_tracing_enabled():
                links = get_span_links_from_received_events(event)
                if isinstance(event, list):
                    is_batch = True

            with receive_context_manager(
                self._eventhub_client, links=links, start_time=self._last_received_time
            ):
                self._last_received_time = time.time_ns()

            with process_context_manager(self._eventhub_client, links=links, is_batch=is_batch):
                self._event_handler(partition_context, event)  # type: ignore
        else:
            self._event_handler(partition_context, event)  # type: ignore

    def _load_balancing(self) -> None:
        """Start the EventProcessor.

        The EventProcessor will try to claim and balance partition ownership with other `EventProcessor`
        and start receiving EventData from EventHub and processing events.

        """
        while self._running:
            random_jitter = self._load_balancing_interval * random.random() * 0.2
            load_balancing_interval = self._load_balancing_interval + random_jitter
            try:
                claimed_partition_ids = self._ownership_manager.claim_ownership()
                if claimed_partition_ids:
                    existing_pids = set(self._consumers.keys())
                    claimed_pids = set(claimed_partition_ids)
                    to_cancel_pids = existing_pids - claimed_pids
                    newly_claimed_pids = claimed_pids - existing_pids
                    if newly_claimed_pids:
                        checkpoints = self._ownership_manager.get_checkpoints() if self._checkpoint_store else None
                        self._create_tasks_for_claimed_ownership(newly_claimed_pids, checkpoints)
                else:
                    _LOGGER.info(
                        "EventProcessor %r hasn't claimed an ownership. It keeps claiming.",
                        self._id,
                    )
                    to_cancel_pids = set(self._consumers.keys())
                if to_cancel_pids:
                    self._cancel_tasks_for_partitions(to_cancel_pids)
            except Exception as err:  # pylint:disable=broad-except
                # ownership_manager.get_checkpoints() and ownership_manager.claim_ownership() may raise exceptions
                # when there are load balancing and/or checkpointing (checkpoint_store isn't None).
                # They're swallowed here to retry every self._load_balancing_interval seconds.
                # Meanwhile this event processor won't lose the partitions it has claimed before.
                # If it keeps failing, other EventProcessors will start to claim ownership of the partitions
                # that this EventProcessor is working on. So two or multiple EventProcessors may be working
                # on the same partition for a short while.
                # Setting owner_level would create exclusive connection to the partition and
                # alleviate duplicate-receiving greatly.
                _LOGGER.warning(
                    "EventProcessor instance %r of eventhub %r consumer group %r. "
                    "An error occurred while load-balancing and claiming ownership. "
                    "The exception is %r. Retrying after %r seconds",
                    self._id,
                    self._eventhub_name,
                    self._consumer_group,
                    err,
                    load_balancing_interval,
                )
                self._process_error(None, err)  # type: ignore

            time.sleep(load_balancing_interval)

    def _close_consumer(self, partition_id: str, consumer: EventHubConsumer, reason: CloseReason) -> None:
        consumer.close()
        with self._lock:
            del self._consumers[partition_id]

        _LOGGER.info(
            "PartitionProcessor of EventProcessor instance %r of eventhub %r partition %r consumer group %r"
            " is being closed. Reason is: %r",
            self._id,
            self._partition_contexts[partition_id].eventhub_name,
            self._partition_contexts[partition_id].partition_id,
            self._partition_contexts[partition_id].consumer_group,
            reason,
        )

        if self._partition_close_handler:
            try:
                self._partition_close_handler(self._partition_contexts[partition_id], reason)
            except Exception as err:  # pylint:disable=broad-except
                _LOGGER.warning(
                    "EventProcessor instance %r of eventhub %r partition %r consumer group %r. "
                    "An error occurred while running on_partition_close. The exception is %r.",
                    self._id,
                    self._partition_contexts[partition_id].eventhub_name,
                    self._partition_contexts[partition_id].partition_id,
                    self._partition_contexts[partition_id].consumer_group,
                    err,
                )
                self._process_error(self._partition_contexts[partition_id], err)

        self._ownership_manager.release_ownership(partition_id)

    def _do_receive(self, partition_id: str, consumer: EventHubConsumer) -> None:
        """Call the consumer.receive() and handle exceptions if any after it exhausts retries.

        :param str partition_id: The partition id.
        :param ~azure.eventhub._consumer.EventHubConsumer consumer: The consumer object.
        """
        try:
            consumer.receive(self._batch, self._max_batch_size, self._max_wait_time)
        except Exception as error:  # pylint:disable=broad-except
            _LOGGER.warning(
                "EventProcessor instance %r of eventhub %r partition %r consumer group %r. "
                "An error occurred while receiving. The exception is %r.",
                self._id,
                self._partition_contexts[partition_id].eventhub_name,
                self._partition_contexts[partition_id].partition_id,
                self._partition_contexts[partition_id].consumer_group,
                error,
            )
            self._process_error(self._partition_contexts[partition_id], error)
            # TODO: close consumer if non-retryable. issue #27137
            # Does OWNERSHIP_LOST make sense for all errors?
            self._close_consumer(partition_id, consumer, CloseReason.OWNERSHIP_LOST)

    def start(self) -> None:
        if self._running:
            _LOGGER.info("EventProcessor %r has already started.", self._id)
            return

        _LOGGER.info("EventProcessor %r is being started", self._id)
        self._running = True
        thread = threading.Thread(target=self._load_balancing)
        thread.daemon = True
        thread.start()

        while self._running:
            for partition_id, consumer in list(self._consumers.items()):
                if consumer.stop:
                    self._close_consumer(partition_id, consumer, CloseReason.OWNERSHIP_LOST)
                    continue

                self._do_receive(partition_id, consumer)

        with self._lock:
            for partition_id, consumer in list(self._consumers.items()):
                self._close_consumer(partition_id, consumer, CloseReason.SHUTDOWN)

    def stop(self) -> None:
        """Stop the EventProcessor.

        The EventProcessor will stop receiving events from EventHubs and release the ownership of the partitions
        it is working on.
        Other running EventProcessor will take over these released partitions.

        A stopped EventProcessor can be restarted by calling method `start` again.

        """
        if not self._running:
            _LOGGER.info("EventProcessor %r has already been stopped.", self._id)
            return

        self._running = False
        _LOGGER.info("EventProcessor %r has been stopped.", self._id)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show how to create async EventHubProducerClient and EventHubConsumerClient that connect to custom endpoint.
"""

import os
import asyncio
from azure.eventhub import EventData
from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]
# The custom endpoint address to use for establishing a connection to the Event Hubs service,
# allowing network requests to be routed through any application gateways
# or other paths needed for the host environment.
CUSTOM_ENDPOINT_ADDRESS = "sb://<custom_endpoint_hostname>:<custom_endpoint_port>"
# The optional absolute path to the custom certificate file used by client to authenticate the
# identity of the connection endpoint in the case that endpoint has its own issued CA.
# If not set, the certifi library will be used to load certificates.
CUSTOM_CA_BUNDLE_PATH = "<your_custom_ca_bundle_file_path>"


async def producer_connecting_to_custom_endpoint():
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        credential=DefaultAzureCredential(),
        eventhub_name=EVENTHUB_NAME,
        custom_endpoint_address=CUSTOM_ENDPOINT_ADDRESS,
        connection_verify=CUSTOM_CA_BUNDLE_PATH,
    )

    async with producer_client:
        # Without specifying partition_id or partition_key
        # the events will be distributed to available partitions via round-robin.
        event_data_batch = await producer_client.create_batch()
        event_data_batch.add(EventData("Single message"))
        await producer_client.send_batch(event_data_batch)
        print("Send a message.")


async def on_event(partition_context, event):
    # Put your code here.
    # If the operation is i/o intensive, multi-thread will have better performance.
    print("Received event from partition: {}.".format(partition_context.partition_id))


async def consumer_connecting_to_custom_endpoint():
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        credential=DefaultAzureCredential(),
        eventhub_name=EVENTHUB_NAME,
        consumer_group="$Default",
        custom_endpoint_address=CUSTOM_ENDPOINT_ADDRESS,
        connection_verify=CUSTOM_CA_BUNDLE_PATH,
    )

    try:
        async with consumer_client:
            await consumer_client.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
    except KeyboardInterrupt:
        print("Stopped receiving.")


asyncio.run(producer_connecting_to_custom_endpoint())
asyncio.run(consumer_connecting_to_custom_endpoint())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.eventhub._pyamqp.message import Message
from azure.eventhub._pyamqp.utils import AMQPTypes
from azure.eventhub._pyamqp._encode import encode_payload
from azure.eventhub._pyamqp._decode import decode_payload
from azure.eventhub.amqp._amqp_message import AmqpMessageProperties


def test_message():
    value = Message(value="test_message")

    assert value.data == None
    assert value.value == "test_message"


def test_body_value():
    message = Message(value="test_message")
    body_value = b"body"

    message = message._replace(data=body_value)
    assert message.data == body_value

    output = bytearray()
    encode_payload(output, message)

    message = decode_payload(memoryview(output))
    output = bytearray()

    body = message.data
    assert body[0].get(b"TYPE").decode("utf-8") == AMQPTypes.binary


def test_delivery_tag():
    message = Message(value="test_message")
    assert not message.delivery_annotations


def test_message_properties():

    value = AmqpMessageProperties()
    assert not value.user_id

    value = AmqpMessageProperties()
    value.user_id = bytearray(b"testuseridlongstring")
    assert value.user_id == b"testuseridlongstring"

    value = AmqpMessageProperties()
    value.user_id = bytearray(b"")
    assert value.user_id == b""

    value = AmqpMessageProperties()
    value.user_id = bytearray(b"short")
    assert value.user_id == b"short"

    value = AmqpMessageProperties()
    value.user_id = bytearray(b"!@#$%^&*()+_?")
    assert value.user_id == b"!@#$%^&*()+_?"

    value = AmqpMessageProperties()
    value.user_id = bytearray(b"\nweird\0user\1id\0\t")
    assert value.user_id == b"\nweird\0user\1id\0\t"



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub with checkpoint store.
In the `receive` method of `EventHubConsumerClient`:
If no partition id is specified, the checkpoint_store are used for load-balance and checkpoint.
If partition id is specified, the checkpoint_store can only be used for checkpoint.
"""
import os
from azure.eventhub import EventHubConsumerClient
from azure.eventhub.extensions.checkpointstoreblob import BlobCheckpointStore
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]

storage_account_name = os.environ["AZURE_STORAGE_ACCOUNT"]
protocol = os.environ.get("PROTOCOL", "https")
suffix = os.environ.get("ACCOUNT_URL_SUFFIX", "core.windows.net")
BLOB_ACCOUNT_URL = f"{protocol}://{storage_account_name}.blob.{suffix}"
BLOB_CONTAINER_NAME = "your-blob-container-name"  # Please make sure the blob container resource exists.


def on_event(partition_context, event):
    # Put your code here.
    # Avoid time-consuming operations.
    print("Received event from partition: {}.".format(partition_context.partition_id))
    partition_context.update_checkpoint(event)


if __name__ == "__main__":
    checkpoint_store = BlobCheckpointStore(
        blob_account_url=BLOB_ACCOUNT_URL, container_name=BLOB_CONTAINER_NAME, credential=DefaultAzureCredential()
    )
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        consumer_group="$Default",
        checkpoint_store=checkpoint_store,  # For load-balancing and checkpoint. Leave None for no load-balancing.
    )

    try:
        with consumer_client:
            """
            Without specified partition_id, the receive will try to receive events from all partitions and if provided
            with a checkpoint store, the client will load-balance partition assignment with other EventHubConsumerClient
            instances which also try to receive events from all partitions and use the same storage resource.
            """
            consumer_client.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
            # With specified partition_id, load-balance will be disabled, for example:
            # client.receive(on_event=on_event, partition_id='0')
    except KeyboardInterrupt:
        print("Stop receiving.")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations

import sys
import platform
import datetime
import calendar
import logging
from typing import TYPE_CHECKING, cast, Type, Optional, Dict, Union, Any, Iterable, Tuple, Mapping, Callable
from datetime import timezone
from .amqp import AmqpAnnotatedMessage, AmqpMessageHeader
from ._version import VERSION
from ._constants import (
    MAX_USER_AGENT_LENGTH,
    USER_AGENT_PREFIX,
    PROP_LAST_ENQUEUED_SEQUENCE_NUMBER,
    PROP_LAST_ENQUEUED_TIME_UTC,
    PROP_RUNTIME_INFO_RETRIEVAL_TIME_UTC,
    PROP_LAST_ENQUEUED_OFFSET,
)

if TYPE_CHECKING:
    from ._transport._base import AmqpTransport
    from ._pyamqp.message import Message as pyamqp_Message

    try:
        from uamqp import types as uamqp_types
        from uamqp import Message as uamqp_Message
    except ImportError:
        uamqp_types = None
    from azure.core.credentials import AzureSasCredential
    from ._common import EventData

    MessagesType = Union[
        AmqpAnnotatedMessage,
        EventData,
        Iterable[Union[AmqpAnnotatedMessage, EventData]],
    ]

_LOGGER = logging.getLogger(__name__)


TZ_UTC: timezone = timezone.utc
# Number of seconds between the Unix epoch (1/1/1970) and year 1 CE.
# This is the lowest value that can be represented by an AMQP timestamp.
CE_ZERO_SECONDS: int = -62_135_596_800


def utc_from_timestamp(timestamp: float) -> datetime.datetime:
    """
    :param float timestamp: Timestamp in seconds to be converted to datetime.
    :rtype: datetime.datetime
    :returns: A datetime object representing the timestamp in UTC.
    """
    # The AMQP timestamp is the number of seconds since the Unix epoch.
    # AMQP brokers represent the lowest value as -62_135_596_800 (the
    # number of seconds between the Unix epoch (1/1/1970) and year 1 CE) as
    # a sentinel for a time which is not set.
    if timestamp == CE_ZERO_SECONDS:
        return datetime.datetime.min.replace(tzinfo=TZ_UTC)
    return datetime.datetime.fromtimestamp(timestamp, tz=TZ_UTC)


def create_properties(
    user_agent: Optional[str] = None, *, amqp_transport: AmqpTransport
) -> Union[Dict[uamqp_types.AMQPSymbol, str], Dict[str, str]]:
    """
    Format the properties with which to instantiate the connection.
    This acts like a user agent over HTTP.

    :param str or None user_agent: The user agent string.
    :keyword ~azure.eventhub._transport._base.AmqpTransport amqp_transport: The AMQP transport.
    :rtype: dict
    :return: The properties.
    """
    properties: Dict[Any, str] = {}
    properties[amqp_transport.PRODUCT_SYMBOL] = USER_AGENT_PREFIX
    properties[amqp_transport.VERSION_SYMBOL] = VERSION
    framework = f"Python/{sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}"
    properties[amqp_transport.FRAMEWORK_SYMBOL] = framework
    platform_str = platform.platform()
    properties[amqp_transport.PLATFORM_SYMBOL] = platform_str

    final_user_agent = f"{USER_AGENT_PREFIX}/{VERSION} {amqp_transport.TRANSPORT_IDENTIFIER} {framework} ({platform_str})"  # pylint: disable=line-too-long
    if user_agent:
        final_user_agent = f"{user_agent} {final_user_agent}"

    if len(final_user_agent) > MAX_USER_AGENT_LENGTH:
        raise ValueError(
            f"The user-agent string cannot be more than {MAX_USER_AGENT_LENGTH} in length."
            f"Current user_agent string is: {final_user_agent} with length: {len(final_user_agent)}"
        )
    properties[amqp_transport.USER_AGENT_SYMBOL] = final_user_agent
    return properties


def set_event_partition_key(
    event: Union[AmqpAnnotatedMessage, EventData],
    partition_key: Optional[Union[bytes, str]],
    amqp_transport: AmqpTransport,
) -> None:
    if not partition_key:
        return

    try:
        raw_message = event.raw_amqp_message  # type: ignore
    except AttributeError:
        raw_message = event

    annotations = raw_message.annotations
    if annotations is None:
        annotations = {}
    annotations[amqp_transport.PROP_PARTITION_KEY_AMQP_SYMBOL] = partition_key
    if not raw_message.header:
        raw_message.header = AmqpMessageHeader(header=True)
    else:
        raw_message.header.durable = True


def event_position_selector(value: Union[str, int, datetime.datetime], inclusive: bool = False) -> bytes:
    """Creates a selector expression of the offset.

    :param int or str or datetime.datetime value: The offset value to use for the offset.
    :param bool inclusive: Whether to include the value in the range.
    :rtype: bytes
    :return: The selector filter expression.
    """
    operator = ">=" if inclusive else ">"
    if isinstance(value, datetime.datetime):  # pylint:disable=no-else-return
        timestamp = (calendar.timegm(value.utctimetuple()) * 1000) + (value.microsecond / 1000)
        return (f"amqp.annotation.x-opt-enqueued-time {operator} '{int(timestamp)}'").encode("utf-8")
    elif isinstance(value, int):
        return (f"amqp.annotation.x-opt-sequence-number {operator} '{value}'").encode("utf-8")
    return (f"amqp.annotation.x-opt-offset {operator} '{value}'").encode("utf-8")


def get_last_enqueued_event_properties(event_data: EventData) -> Optional[Dict[str, Any]]:
    """Extracts the last enqueued event in from the received event delivery annotations.

    :param ~azure.eventhub.EventData event_data: The received Event Data.
    :rtype: dict[str, any] or None
    :return: The enqueued event properties dictionary.
    """
    # pylint: disable=protected-access
    if event_data._last_enqueued_event_properties:
        return event_data._last_enqueued_event_properties

    if event_data._message.delivery_annotations:
        sequence_number = event_data._message.delivery_annotations.get(PROP_LAST_ENQUEUED_SEQUENCE_NUMBER, None)
        enqueued_time_stamp = event_data._message.delivery_annotations.get(PROP_LAST_ENQUEUED_TIME_UTC, None)
        if enqueued_time_stamp:
            enqueued_time_stamp = utc_from_timestamp(float(enqueued_time_stamp) / 1000)
        retrieval_time_stamp = event_data._message.delivery_annotations.get(PROP_RUNTIME_INFO_RETRIEVAL_TIME_UTC, None)
        if retrieval_time_stamp:
            retrieval_time_stamp = utc_from_timestamp(float(retrieval_time_stamp) / 1000)
        offset_bytes = event_data._message.delivery_annotations.get(PROP_LAST_ENQUEUED_OFFSET, None)
        offset = offset_bytes.decode("UTF-8") if offset_bytes else None

        event_data._last_enqueued_event_properties = {
            "sequence_number": sequence_number,
            "offset": offset,
            "enqueued_time": enqueued_time_stamp,
            "retrieval_time": retrieval_time_stamp,
        }
        return event_data._last_enqueued_event_properties
    return None


def parse_sas_credential(credential: AzureSasCredential) -> Tuple[str, Optional[int]]:
    sas = credential.signature
    parsed_sas = sas.split("&")
    expiry = None
    for item in parsed_sas:
        if item.startswith("se="):
            expiry = int(item[3:])
    return (sas, expiry)


def transform_outbound_single_message(
    message: Union[AmqpAnnotatedMessage, EventData],
    message_type: Type[EventData],
    to_outgoing_amqp_message: Callable[[AmqpAnnotatedMessage], Union[uamqp_Message, pyamqp_Message]],
) -> EventData:
    """
    This method serves multiple goals:
    1. update the internal message to reflect any updates to settable properties on EventData
    2. transform the AmqpAnnotatedMessage to be EventData
    :param message: A single instance of message of type EventData
        or AmqpAnnotatedMessage.
    :type message: ~azure.eventhub.EventData or ~azure.eventhub.amqp.AmqpAnnotatedMessage
    :param type[~azure.eventhub.EventData] message_type: The class type to return the messages as.
    :param callable to_outgoing_amqp_message: A function to transform the message
    :rtype:  ~azure.eventhub.EventData
    :return: The transformed message.
    """
    try:
        # pylint: disable=protected-access
        # If EventData, set EventData._message to uamqp/pyamqp.Message right before sending.
        message = cast("EventData", message)
        message._message = to_outgoing_amqp_message(message.raw_amqp_message)
        return message  # type: ignore
    except AttributeError:
        # pylint: disable=protected-access
        # If AmqpAnnotatedMessage, create EventData object with _from_message.
        # event_data._message will be set to outgoing uamqp/pyamqp.Message.
        # event_data.raw_amqp_message will be set to AmqpAnnotatedMessage.
        message = cast(AmqpAnnotatedMessage, message)
        amqp_message = to_outgoing_amqp_message(message)
        return message_type._from_message(message=amqp_message, raw_amqp_message=message)  # type: ignore


def decode_with_recurse(data: Any, encoding: str = "UTF-8") -> Any:
    """
    If data is of a compatible type, iterates through nested structure and decodes all binary
        strings with provided encoding.
    :param any data: The data object which, if compatible, will be iterated through to decode binary string.
    :param str encoding: The encoding to use for decoding data.
        Default is 'UTF-8'
    :rtype: any
    :return: The decoded data object.
    """

    if isinstance(data, str):
        return data
    if isinstance(data, bytes):
        return data.decode(encoding)
    if isinstance(data, Mapping):
        decoded_mapping = {}
        for k, v in data.items():
            decoded_key = decode_with_recurse(k, encoding)
            decoded_val = decode_with_recurse(v, encoding)
            decoded_mapping[decoded_key] = decoded_val
        return decoded_mapping
    if isinstance(data, Iterable):
        decoded_list = []
        for d in data:
            decoded_list.append(decode_with_recurse(d, encoding))
        return decoded_list

    return data



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show sending events with different options to an Event Hub partition.

WARNING: EventHubProducerClient and EventDataBatch are not thread-safe.
Do not share these instances between threads without proper thread-safe management using mechanisms like threading.Lock.
Note: Native async APIs should be used instead of running in a ThreadPoolExecutor, if possible.
"""

import time
import os
import threading
from concurrent.futures import ThreadPoolExecutor

from azure.eventhub import EventHubProducerClient, EventData
from azure.eventhub.exceptions import EventHubError

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


# [START send_event_data_batch]
def send_event_data_batch(producer):
    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.
    event_data_batch = producer.create_batch()
    event_data_batch.add(EventData("Single message"))
    producer.send_batch(event_data_batch)


# [END send_event_data_batch]


def send_event_data_batch_with_limited_size(producer):
    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.
    event_data_batch_with_limited_size = producer.create_batch(max_size_in_bytes=1000)

    while True:
        try:
            event_data_batch_with_limited_size.add(EventData("Message inside EventBatchData"))
        except ValueError:
            # EventDataBatch object reaches max_size.
            # New EventDataBatch object can be created here to send more data.
            break

    producer.send_batch(event_data_batch_with_limited_size)


def send_event_data_batch_with_partition_key(producer):
    # Specifying partition_key.
    event_data_batch_with_partition_key = producer.create_batch(partition_key="pkey")
    event_data_batch_with_partition_key.add(
        EventData("Message will be sent to a partition determined by the partition key")
    )

    producer.send_batch(event_data_batch_with_partition_key)


def send_event_data_batch_with_partition_id(producer):
    # Specifying partition_id.
    event_data_batch_with_partition_id = producer.create_batch(partition_id="0")
    event_data_batch_with_partition_id.add(EventData("Message will be sent to target-id partition"))

    producer.send_batch(event_data_batch_with_partition_id)


def send_event_data_batch_with_properties(producer):
    event_data_batch = producer.create_batch()
    event_data = EventData("Message with properties")
    event_data.properties = {"prop_key": "prop_value"}
    event_data_batch.add(event_data)
    producer.send_batch(event_data_batch)


def send_event_data_list(producer):
    # If you know beforehand that the list of events you have will not exceed the
    # size limits, you can use the `send_batch()` api directly without creating an EventDataBatch

    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.

    event_data_list = [EventData("Event Data {}".format(i)) for i in range(10)]
    try:
        producer.send_batch(event_data_list)
    except ValueError:  # Size exceeds limit. This shouldn't happen if you make sure before hand.
        print("Size of the event data list exceeds the size limit of a single send")
    except EventHubError as eh_err:
        print("Sending error: ", eh_err)


def send_concurrent_with_shared_client_and_lock():
    """
    Example showing concurrent sending with a shared client using threading.Lock.
    Note: Native async APIs should be used instead of running in a ThreadPoolExecutor, if possible.
    """
    send_lock = threading.Lock()
    
    producer = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        eventhub_name=EVENTHUB_NAME,
    )

    def send_with_lock(thread_id):
        try:
            # Use lock to ensure thread-safe sending
            with send_lock:
                batch = producer.create_batch()
                batch.add(EventData(f"Synchronized message from thread {thread_id}"))
                producer.send_batch(batch)
                print(f"Thread {thread_id} sent synchronized message successfully")
        except Exception as e:
            print(f"Thread {thread_id} failed: {e}")

    with producer:
        with ThreadPoolExecutor(max_workers=3) as executor:
            futures = [executor.submit(send_with_lock, i) for i in range(3)]
            # Wait for all threads to complete
            for future in futures:
                future.result()


producer = EventHubProducerClient.from_connection_string(
    conn_str=CONNECTION_STR,
    eventhub_name=EVENTHUB_NAME,
)

start_time = time.time()
with producer:
    send_event_data_batch(producer)
    send_event_data_batch_with_limited_size(producer)
    send_event_data_batch_with_partition_key(producer)
    send_event_data_batch_with_partition_id(producer)
    send_event_data_batch_with_properties(producer)
    send_event_data_list(producer)

print("Send messages in {} seconds.".format(time.time() - start_time))


print("\nDemonstrating concurrent sending with shared client and locks...")
send_concurrent_with_shared_client_and_lock()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# -----------------------------------------------------------------------------------
import math
import time
import random
from collections import Counter, defaultdict
from typing import List, Iterable, Optional, Dict, Any, Union, TYPE_CHECKING

from ..._eventprocessor.common import LoadBalancingStrategy

if TYPE_CHECKING:
    from .checkpoint_store import CheckpointStore
    from .._consumer_client_async import EventHubConsumerClient
    from .._producer_client_async import EventHubProducerClient


class OwnershipManager:  # pylint:disable=too-many-instance-attributes
    """Increases or decreases the number of partitions owned by an EventProcessor
    so the number of owned partitions are balanced among multiple EventProcessors

    An EventProcessor calls claim_ownership() of this class every x seconds,
    where x is set by keyword argument "load_balancing_interval" in EventProcessor,
    to claim the ownership of partitions, create tasks for the claimed ownership, and cancel tasks that no longer belong
    to the claimed ownership.

    """

    def __init__(
        self,
        eventhub_client: Union["EventHubConsumerClient", "EventHubProducerClient"],
        consumer_group: str,
        owner_id: str,
        checkpoint_store: Optional["CheckpointStore"],
        ownership_timeout: float,
        load_balancing_strategy: LoadBalancingStrategy,
        partition_id: Optional[str],
    ):
        self.cached_parition_ids: List[str] = []
        self.owned_partitions: Iterable[Dict[str, Any]] = []
        self.eventhub_client = eventhub_client
        self.fully_qualified_namespace = eventhub_client._address.hostname
        self.eventhub_name = eventhub_client.eventhub_name
        self.consumer_group = consumer_group
        self.owner_id = owner_id
        self.checkpoint_store = checkpoint_store
        self.ownership_timeout = ownership_timeout
        self.load_balancing_strategy = load_balancing_strategy
        self.partition_id = partition_id

    async def claim_ownership(self) -> List[str]:
        """Claims ownership for this EventProcessor

        :return: The list of partition ids claimed by this EventProcessor.
        :rtype: list[str]
        """
        if not self.cached_parition_ids:
            await self._retrieve_partition_ids()

        if self.partition_id is not None:
            if self.partition_id in self.cached_parition_ids:
                return [self.partition_id]
            raise ValueError(
                "Wrong partition id:{}. The eventhub has partitions: {}.".format(
                    self.partition_id, self.cached_parition_ids
                )
            )

        if self.checkpoint_store is None:
            return self.cached_parition_ids

        ownership_list = await self.checkpoint_store.list_ownership(
            self.fully_qualified_namespace, self.eventhub_name, self.consumer_group
        )
        to_claim = self._balance_ownership(ownership_list, self.cached_parition_ids)
        self.owned_partitions = await self.checkpoint_store.claim_ownership(to_claim) if to_claim else []
        return [x["partition_id"] for x in self.owned_partitions]

    async def release_ownership(self, partition_id: str) -> None:
        """Explicitly release ownership of a partition if we still have it.

        This is called when a consumer is shutdown, and is achieved by resetting the associated
        owner ID.

        :param partition_id: The partition id to release ownership.
        :type partition_id: str
        """
        if not self.checkpoint_store:
            return
        now = time.time()
        partition_ownership = [
            o
            for o in self.owned_partitions
            if o["partition_id"] == partition_id
            and o["owner_id"] == self.owner_id
            and o["last_modified_time"] + self.ownership_timeout > now
        ]
        if not partition_ownership:
            return
        partition_ownership[0]["owner_id"] = ""
        await self.checkpoint_store.claim_ownership(partition_ownership)

    async def _retrieve_partition_ids(self) -> None:
        """List all partition ids of the event hub that the EventProcessor is working on."""
        self.cached_parition_ids = await self.eventhub_client.get_partition_ids()

    def _balance_ownership(  # pylint:disable=too-many-locals
        self, ownership_list: Iterable[Dict[str, Any]], all_partition_ids: List[str]
    ) -> List[Dict[str, Any]]:
        """Balances and claims ownership of partitions for this EventProcessor.

        :param ownership_list: The list of current ownership.
        :type ownership_list: iterable[dict[str, any]]
        :param all_partition_ids: The list of all partition ids.
        :type all_partition_ids: list[str]
        :return: The list of partition ownership that this EventProcessor claims.
        :rtype: list[dict[str, any]]
        """
        now = time.time()
        ownership_dict = {x["partition_id"]: x for x in ownership_list}  # put the list to dict for fast lookup
        unclaimed_partition_ids = [pid for pid in all_partition_ids if pid not in ownership_dict]
        released_partitions = [
            x for x in ownership_list if x["last_modified_time"] + self.ownership_timeout < now or not x["owner_id"]
        ]
        released_partition_ids = [ownership["partition_id"] for ownership in released_partitions]
        claimable_partition_ids = unclaimed_partition_ids + released_partition_ids

        active_ownership = [o for o in ownership_list if o not in released_partitions]
        active_ownership_by_owner: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
        for ownership in active_ownership:
            active_ownership_by_owner[ownership["owner_id"]].append(ownership)
        active_ownership_self = active_ownership_by_owner[self.owner_id]

        # calculate expected and max count per owner
        all_partition_count = len(all_partition_ids)
        # owners_count is the number of active owners. If self.owner_id is not yet among the active owners,
        # then plus 1 to include self. This will make owners_count >= 1.
        owners_count = len(active_ownership_by_owner) + (0 if self.owner_id in active_ownership_by_owner else 1)
        expected_count_per_owner = all_partition_count // owners_count

        # Py2 math.ceil() returns float, a/b return int if not divisable.
        # Py3 math.ceil() returns int, a/b return float if not divisable.
        # Even though this is py3 code, make it the same as the py2-compatible sync code.
        max_count_per_owner = int(math.ceil(all_partition_count * 1.0 / owners_count))
        # end of calculating expected count per owner

        to_claim = active_ownership_self
        if len(active_ownership_self) < max_count_per_owner:
            to_try_steal = True
            if self.load_balancing_strategy is LoadBalancingStrategy.GREEDY:
                # Greedily claim more partitions if there are claimable partitions
                to_greedy_claim_ids = random.sample(
                    claimable_partition_ids,
                    k=min(
                        max_count_per_owner - len(active_ownership_self),
                        len(claimable_partition_ids),
                    ),
                )
                if to_greedy_claim_ids:
                    for pid in to_greedy_claim_ids:
                        random_chosen_to_claim = ownership_dict.get(
                            pid,
                            {
                                "fully_qualified_namespace": self.fully_qualified_namespace,
                                "partition_id": pid,
                                "eventhub_name": self.eventhub_name,
                                "consumer_group": self.consumer_group,
                            },
                        )
                        random_chosen_to_claim["owner_id"] = self.owner_id
                        to_claim.append(random_chosen_to_claim)
                    to_try_steal = False  # already greedily got at least one
            else:  # self.load_balancing_strategy is LoadBalancingStrategy.BALANCED
                if claimable_partition_ids:  # claim an inactive partition if there is
                    random_partition_id = random.choice(claimable_partition_ids)
                    random_chosen_to_claim = ownership_dict.get(
                        random_partition_id,
                        {
                            "fully_qualified_namespace": self.fully_qualified_namespace,
                            "partition_id": random_partition_id,
                            "eventhub_name": self.eventhub_name,
                            "consumer_group": self.consumer_group,
                        },
                    )
                    random_chosen_to_claim["owner_id"] = self.owner_id
                    to_claim.append(random_chosen_to_claim)
                    to_try_steal = False  # already got one from claimable partition.
            if to_try_steal and len(active_ownership_self) < expected_count_per_owner:
                # steal from another owner that has the most count
                active_ownership_count_group_by_owner = Counter(
                    dict((x, len(y)) for x, y in active_ownership_by_owner.items())
                )
                most_frequent_owner_id = active_ownership_count_group_by_owner.most_common(1)[0][0]
                # randomly choose a partition to steal from the most_frequent_owner
                to_steal_partition = random.choice(active_ownership_by_owner[most_frequent_owner_id])
                to_steal_partition["owner_id"] = self.owner_id
                to_claim.append(to_steal_partition)
        return to_claim

    async def get_checkpoints(self) -> Dict[str, Dict[str, Any]]:
        if self.checkpoint_store:
            checkpoints = await self.checkpoint_store.list_checkpoints(
                self.fully_qualified_namespace, self.eventhub_name, self.consumer_group
            )
            return {x["partition_id"]: x for x in checkpoints}
        return {}



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub with checkpoint store doing checkpoint by batch asynchronously.
In the `receive_batch` method of `EventHubConsumerClient`:
If no partition id is specified, the checkpoint_store are used for load-balance and checkpoint.
If partition id is specified, the checkpoint_store can only be used for checkpoint without load balancing.
"""

import asyncio
import os
import logging
from azure.eventhub.aio import EventHubConsumerClient
from azure.eventhub.extensions.checkpointstoreblobaio import BlobCheckpointStore
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]

storage_account_name = os.environ["AZURE_STORAGE_ACCOUNT"]
protocol = os.environ.get("PROTOCOL", "https")
suffix = os.environ.get("ACCOUNT_URL_SUFFIX", "core.windows.net")
BLOB_ACCOUNT_URL = f"{protocol}://{storage_account_name}.blob.{suffix}"
BLOB_CONTAINER_NAME = "your-blob-container-name"  # Please make sure the blob container resource exists.

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


async def batch_process_events(events):
    # put your code here
    await asyncio.sleep(2)  # simulate something I/O bound


async def on_event_batch(partition_context, event_batch):
    log.info("Partition {}, Received count: {}".format(partition_context.partition_id, len(event_batch)))
    await batch_process_events(event_batch)
    await partition_context.update_checkpoint()


async def receive_batch():
    checkpoint_store = BlobCheckpointStore(
        blob_account_url=BLOB_ACCOUNT_URL, container_name=BLOB_CONTAINER_NAME, credential=DefaultAzureCredential()
    )
    client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        consumer_group="$Default",
        checkpoint_store=checkpoint_store,
    )
    async with client:
        await client.receive_batch(
            on_event_batch=on_event_batch,
            max_batch_size=100,
            starting_position="-1",  # "-1" is from the beginning of the partition.
        )


if __name__ == "__main__":
    asyncio.run(receive_batch())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
jenkins-hash lookup3 algorithm implementation
"""

from asyncio import Lock  # pylint:disable=do-not-import-asyncio
from ..._buffered_producer._partition_resolver import (
    generate_hash_code,
)


class PartitionResolver:
    def __init__(self, partitions):
        self._idx = -1
        self._partitions = partitions
        self._partitions_cnt = len(self._partitions)
        self._lock = Lock()

    async def get_next_partition_id(self):
        """
        round-robin partition assignment

        :return: The next partition id.
        :rtype: str
        """
        async with self._lock:
            self._idx += 1
            self._idx %= self._partitions_cnt
            return self._partitions[self._idx]

    async def get_partition_id_by_partition_key(self, partition_key):
        hash_code = generate_hash_code(partition_key)
        return self._partitions[abs(hash_code % self._partitions_cnt)]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


import asyncio
import pytest
import time

from azure.eventhub._pyamqp.aio._authentication_async import SASTokenAuthAsync
from azure.eventhub._pyamqp.aio import ReceiveClientAsync
from azure.eventhub._pyamqp import error, constants
from azure.eventhub._utils import transform_outbound_single_message

try:
    import uamqp
    from uamqp import compat
    from azure.eventhub._transport._uamqp_transport import UamqpTransport
except (ModuleNotFoundError, ImportError):
    uamqp = None
    UamqpTransport = None

from azure.eventhub._transport._pyamqp_transport import PyamqpTransport
from azure.eventhub import EventData
from azure.eventhub.aio import (
    EventHubProducerClient,
    EventHubConsumerClient,
    EventHubSharedKeyCredential,
)
from azure.eventhub.exceptions import OperationTimeoutError


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_with_long_interval_async(live_eventhub, sleep, uamqp_transport, timeout_factor, client_args):
    test_partition = "0"
    sender = EventHubProducerClient(
        live_eventhub["hostname"],
        live_eventhub["event_hub"],
        EventHubSharedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"]),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with sender:
        batch = await sender.create_batch(partition_id=test_partition)
        batch.add(EventData(b"A single event"))
        await sender.send_batch(batch)

        if sleep:
            await asyncio.sleep(250)  # EH server side idle timeout is 240 second
        else:
            if uamqp_transport:
                await sender._producers[test_partition]._handler._connection._conn.destroy()
            else:
                await sender._producers[test_partition]._handler._connection.close()
        batch = await sender.create_batch(partition_id=test_partition)
        batch.add(EventData(b"A single event"))
        await sender.send_batch(batch)

    received = []
    uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
    source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
        live_eventhub["hostname"],
        live_eventhub["event_hub"],
        live_eventhub["consumer_group"],
        test_partition,
    )
    if uamqp_transport:
        sas_auth = uamqp.authentication.SASTokenAsync.from_shared_access_key(
            uri, live_eventhub["key_name"], live_eventhub["access_key"]
        )
        receiver = uamqp.async_ops.client_async.ReceiveClientAsync(
            source, auth=sas_auth, debug=False, timeout=5000, prefetch=500
        )
    else:
        sas_auth = SASTokenAuthAsync(uri, uri, live_eventhub["key_name"], live_eventhub["access_key"])
        receiver = ReceiveClientAsync(
            live_eventhub["hostname"],
            source,
            auth=sas_auth,
            debug=False,
            link_credit=500,
        )
    try:
        await receiver.open_async()

        # receive_message_batch() returns immediately once it receives any messages before the max_batch_size
        # and timeout reach. Could be 1, 2, or any number between 1 and max_batch_size.
        # So call it twice to ensure the two events are received.
        received.extend(
            [
                EventData._from_message(x)
                for x in (await receiver.receive_message_batch_async(max_batch_size=1, timeout=5 * timeout_factor))
            ]
        )
        received.extend(
            [
                EventData._from_message(x)
                for x in (await receiver.receive_message_batch_async(max_batch_size=1, timeout=5 * timeout_factor))
            ]
        )
    finally:
        await receiver.close_async()
    assert len(received) == 2
    assert list(received[0].body)[0] == b"A single event"


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_connection_idle_timeout_and_reconnect_async(
    auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args
):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    if uamqp_transport:
        amqp_transport = UamqpTransport
        retry_total = 3
        timeout_exc = compat.TimeoutException
    else:
        amqp_transport = PyamqpTransport
        retry_total = 0
        timeout_exc = TimeoutError

    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        idle_timeout=10,
        retry_total=retry_total,
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        ed = EventData("data")
        sender = client._create_producer(partition_id="0")
        async with sender:
            await sender._open_with_retry()
            await asyncio.sleep(11)
            ed = transform_outbound_single_message(ed, EventData, amqp_transport.to_outgoing_amqp_message)
            sender._unsent_events = [ed._message]
            if uamqp_transport:
                sender._unsent_events[0].on_send_complete = sender._on_outcome
                with pytest.raises(
                    (
                        uamqp.errors.ConnectionClose,
                        uamqp.errors.MessageHandlerError,
                        OperationTimeoutError,
                    )
                ):
                    await sender._send_event_data()
            else:
                with pytest.raises(error.AMQPConnectionError):
                    await sender._send_event_data()
            if uamqp_transport:
                await sender._send_event_data_with_retry()

    # with retry, should work
    if not uamqp_transport:
        client = EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            idle_timeout=10,
        )
        async with client:
            ed = EventData("data")
            sender = client._create_producer(partition_id="0")
        async with sender:
            await sender._open_with_retry()
            await asyncio.sleep(11)
            ed = transform_outbound_single_message(ed, EventData, amqp_transport.to_outgoing_amqp_message)
            sender._unsent_events = [ed._message]
            with pytest.raises(error.AMQPConnectionError):
                await sender._send_event_data()


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_connection_idle_timeout_and_reconnect_async(auth_credential_senders_async, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders_async
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        idle_timeout=10,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async def on_event_received(event):
        on_event_received.event = event

    async with client:
        consumer = client._create_consumer("$default", "0", "-1", on_event_received)
        async with consumer:
            if uamqp_transport:
                await consumer._open_with_retry()
            else:
                await consumer._open()

            await asyncio.sleep(11)
            ed = EventData("Event")
            senders[0].send(ed)

            if uamqp_transport:
                await consumer._handler.do_work_async()
                assert consumer._handler._connection._state == uamqp.c_uamqp.ConnectionState.DISCARDING
                await consumer.receive(batch=False, max_batch_size=1, max_wait_time=10)
            else:
                with pytest.raises(error.AMQPConnectionError):
                    await consumer._handler.do_work_async()
                assert consumer._handler._connection.state == constants.ConnectionState.END
                try:
                    await asyncio.wait_for(consumer.receive(), timeout=10)
                except asyncio.TimeoutError:
                    pass

            assert on_event_received.event.body_as_str() == "Event"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
import asyncio
import time
import ssl
import certifi

from azure.core.credentials import AzureSasCredential, AzureNamedKeyCredential
from azure.identity.aio import EnvironmentCredential
from azure.eventhub import EventData
from azure.eventhub.exceptions import ConnectError
from azure.eventhub.aio import (
    EventHubConsumerClient,
    EventHubProducerClient,
    EventHubSharedKeyCredential,
)
from azure.eventhub.aio._client_base_async import EventHubSASTokenCredential


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_client_token_credential_async(live_eventhub, get_credential_async, uamqp_transport, client_args):
    credential = get_credential_async()
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
        user_agent="customized information",
        auth_timeout=30,
        uamqp_transport=uamqp_transport,
        **client_args
    )
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        consumer_group="$default",
        credential=credential,
        user_agent="customized information",
        auth_timeout=30,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer_client:
        batch = await producer_client.create_batch(partition_id="0")
        batch.add(EventData(body="A single message"))
        await producer_client.send_batch(batch)

    async def on_event(partition_context, event):
        on_event.called = True
        on_event.partition_id = partition_context.partition_id
        on_event.event = event

    on_event.called = False
    async with consumer_client:
        task = asyncio.ensure_future(consumer_client.receive(on_event, partition_id="0", starting_position="-1"))
        await asyncio.sleep(15)
    await task
    assert on_event.called is True
    assert on_event.partition_id == "0"
    assert list(on_event.event.body)[0] == "A single message".encode("utf-8")


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_client_sas_credential_async(live_eventhub, uamqp_transport, client_args):
    # This should "just work" to validate known-good.
    hostname = live_eventhub["hostname"]
    producer_client = EventHubProducerClient.from_connection_string(
        live_eventhub["connection_str"],
        eventhub_name=live_eventhub["event_hub"],
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer_client:
        batch = await producer_client.create_batch(partition_id="0")
        batch.add(EventData(body="A single message"))
        await producer_client.send_batch(batch)

    # This should also work, but now using SAS tokens.
    credential = EventHubSharedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    auth_uri = "sb://{}/{}".format(hostname, live_eventhub["event_hub"])
    token = (await credential.get_token(auth_uri)).token
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=hostname,
        eventhub_name=live_eventhub["event_hub"],
        credential=EventHubSASTokenCredential(token, time.time() + 3000),
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer_client:
        batch = await producer_client.create_batch(partition_id="0")
        batch.add(EventData(body="A single message"))
        await producer_client.send_batch(batch)

    # Finally let's do it with SAS token + conn str
    token_conn_str = "Endpoint=sb://{}/;SharedAccessSignature={};".format(hostname, token)
    conn_str_producer_client = EventHubProducerClient.from_connection_string(
        token_conn_str,
        eventhub_name=live_eventhub["event_hub"],
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with conn_str_producer_client:
        batch = await conn_str_producer_client.create_batch(partition_id="0")
        batch.add(EventData(body="A single message"))
        await conn_str_producer_client.send_batch(batch)


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_client_azure_sas_credential_async(live_eventhub, uamqp_transport, client_args):
    # This should "just work" to validate known-good.
    hostname = live_eventhub["hostname"]
    producer_client = EventHubProducerClient.from_connection_string(
        live_eventhub["connection_str"],
        eventhub_name=live_eventhub["event_hub"],
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer_client:
        batch = await producer_client.create_batch(partition_id="0")
        batch.add(EventData(body="A single message"))
        await producer_client.send_batch(batch)

    credential = EventHubSharedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    auth_uri = "sb://{}/{}".format(hostname, live_eventhub["event_hub"])
    token = (await credential.get_token(auth_uri)).token
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=hostname,
        eventhub_name=live_eventhub["event_hub"],
        auth_timeout=30,
        credential=AzureSasCredential(token),
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer_client:
        batch = await producer_client.create_batch(partition_id="0")
        batch.add(EventData(body="A single message"))
        await producer_client.send_batch(batch)

    assert (await producer_client.get_eventhub_properties()) is not None


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_client_azure_named_key_credential_async(live_eventhub, uamqp_transport, client_args):

    credential = AzureNamedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        consumer_group="$default",
        credential=credential,
        auth_timeout=30,
        user_agent="customized information",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    assert (await consumer_client.get_eventhub_properties()) is not None

    credential.update("foo", "bar")

    with pytest.raises(Exception):
        await consumer_client.get_eventhub_properties()

    credential.update(live_eventhub["key_name"], live_eventhub["access_key"])
    assert (await consumer_client.get_eventhub_properties()) is not None


# New feature only for Pure Python AMQP, not uamqp.
@pytest.mark.liveTest
@pytest.mark.asyncio
@pytest.mark.no_amqpproxy # testing ssl_context
async def test_client_with_ssl_context_async(auth_credentials_async, socket_transport):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async

    # Check that SSLContext with invalid/nonexistent cert file raises an error
    context = ssl.SSLContext(cafile="fakecert.pem")
    context.verify_mode = ssl.CERT_REQUIRED

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        transport_type=socket_transport,
        ssl_context=context,
        retry_total=0,
    )
    async with producer:
        with pytest.raises(ConnectError):
            batch = await producer.create_batch()

    async def on_event(partition_context, event):
        on_event.called = True
        on_event.partition_id = partition_context.partition_id
        on_event.event = event

    async def on_error(partition_context, error):
        on_error.error = error
        await consumer.close()

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        transport_type=socket_transport,
        ssl_context=context,
        retry_total=0,
    )
    on_error.error = None
    async with consumer:
        task = asyncio.ensure_future(consumer.receive(on_event, on_error=on_error, starting_position="-1"))
        await asyncio.sleep(15)
    await task
    assert isinstance(on_error.error, ConnectError)

    # Check that SSLContext with valid cert file doesn't raise an error
    async def verify_context_async():
        # asyncio.to_thread only available in Python 3.9+
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        await asyncio.to_thread(context.load_verify_locations, certifi.where())
        purpose = ssl.Purpose.SERVER_AUTH
        await asyncio.to_thread(context.load_default_certs, purpose=purpose)
        return context

    def verify_context():  # for Python 3.8
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.load_verify_locations(certifi.where())
        purpose = ssl.Purpose.SERVER_AUTH
        context.load_default_certs(purpose=purpose)
        return context

    if hasattr(asyncio, "to_thread"):
        context = await verify_context_async()
    else:
        context = verify_context()

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        transport_type=socket_transport,
        ssl_context=context,
    )
    async with producer:
        batch = await producer.create_batch()
        batch.add(EventData(body="A single message"))
        batch.add(EventData(body="A second message"))
        await producer.send_batch(batch)

    async def on_event(partition_context, event):
        on_event.total += 1

    async def on_error(partition_context, error):
        on_error.error = error
        await consumer.close()

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        transport_type=socket_transport,
        ssl_context=context,
    )
    on_event.total = 0
    on_error.error = None

    async with consumer:
        task = asyncio.ensure_future(consumer.receive(on_event, on_error=on_error, starting_position="-1"))
        await asyncio.sleep(15)
    await task
    assert on_event.total == 2
    assert on_error.error is None



# -------------------------------------------------------------------------  # pylint: disable=file-needs-copyright-header,useless-suppression
# This is a fork of the transport.py which was originally written by Barry Pederson and
# maintained by the Celery project: https://github.com/celery/py-amqp.
#
# Copyright (C) 2009 Barry Pederson <bp@barryp.org>
#
# The license text can also be found here:
# http://www.opensource.org/licenses/BSD-3-Clause
#
# License
# =======
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# -------------------------------------------------------------------------

import asyncio # pylint:disable=do-not-import-asyncio
import errno
import socket
import ssl
import struct
from ssl import SSLError
from io import BytesIO
import logging


from .._platform import KNOWN_TCP_OPTS, SOL_TCP
from .._encode import encode_frame
from .._decode import decode_frame, decode_empty_frame
from ..constants import (
    DEFAULT_WEBSOCKET_HEARTBEAT_SECONDS,
    TLS_HEADER_FRAME,
    WEBSOCKET_PORT,
    AMQP_WS_SUBPROTOCOL,
    CONNECT_TIMEOUT,
)
from .._transport import (
    AMQP_FRAME,
    get_errno,
    to_host_port,
    DEFAULT_SOCKET_SETTINGS,
    SIGNED_INT_MAX,
    _UNAVAIL,
    AMQP_PORT,
)
from ..error import AuthenticationException, ErrorCondition


_LOGGER = logging.getLogger(__name__)


class AsyncTransportMixin:
    async def receive_frame(self, timeout=None, **kwargs):
        try:
            header, channel, payload = await asyncio.wait_for(self.read(**kwargs), timeout=timeout)
            if not payload:
                decoded = decode_empty_frame(header)
            else:
                decoded = decode_frame(payload)
            return channel, decoded
        except (
            TimeoutError,
            socket.timeout,
            asyncio.IncompleteReadError,
            asyncio.TimeoutError,
        ):
            return None, None

    async def read(self, verify_frame_type=0):
        async with self.socket_lock:
            read_frame_buffer = BytesIO()
            try:
                frame_header = memoryview(bytearray(8))
                read_frame_buffer.write(await self._read(8, buffer=frame_header, initial=True))

                channel = struct.unpack(">H", frame_header[6:])[0]
                size = frame_header[0:4]
                if size == AMQP_FRAME:  # Empty frame or AMQP header negotiation
                    return frame_header, channel, None
                size = struct.unpack(">I", size)[0]
                offset = frame_header[4]
                frame_type = frame_header[5]
                if verify_frame_type is not None and frame_type != verify_frame_type:
                    _LOGGER.debug(
                        "Received invalid frame type: %r, expected: %r",
                        frame_type,
                        verify_frame_type,
                        extra=self.network_trace_params,
                    )
                    raise ValueError(f"Received invalid frame type: {frame_type}, expected: {verify_frame_type}")
                # >I is an unsigned int, but the argument to sock.recv is signed,
                # so we know the size can be at most 2 * SIGNED_INT_MAX
                payload_size = size - len(frame_header)
                payload = memoryview(bytearray(payload_size))
                if size > SIGNED_INT_MAX:
                    read_frame_buffer.write(await self._read(SIGNED_INT_MAX, buffer=payload))
                    read_frame_buffer.write(await self._read(size - SIGNED_INT_MAX, buffer=payload[SIGNED_INT_MAX:]))
                else:
                    read_frame_buffer.write(await self._read(payload_size, buffer=payload))
            except (
                asyncio.CancelledError,
                asyncio.TimeoutError,
                TimeoutError,
                socket.timeout,
                asyncio.IncompleteReadError,
            ):
                read_frame_buffer.write(self._read_buffer.getvalue())
                self._read_buffer = read_frame_buffer
                self._read_buffer.seek(0)
                raise
            except (OSError, IOError, SSLError, socket.error) as exc:
                # Don't disconnect for ssl read time outs
                # http://bugs.python.org/issue10272
                if isinstance(exc, SSLError) and "timed out" in str(exc):
                    raise socket.timeout()
                if get_errno(exc) not in _UNAVAIL:
                    self.connected = False
                _LOGGER.debug("Transport read failed: %r", exc, extra=self.network_trace_params)
                raise
            offset -= 2
        return frame_header, channel, payload[offset:]

    async def write(self, s):
        async with self.socket_lock:
            try:
                await self._write(s)
            except socket.timeout:
                raise
            except (OSError, IOError, socket.error) as exc:
                _LOGGER.debug("Transport write failed: %r", exc, extra=self.network_trace_params)
                if get_errno(exc) not in _UNAVAIL:
                    self.connected = False
                raise

    async def send_frame(self, channel, frame, **kwargs):
        header, performative = encode_frame(frame, **kwargs)
        if performative is None:
            data = header
        else:
            encoded_channel = struct.pack(">H", channel)
            data = header + encoded_channel + performative

        await self.write(data)

    def _build_ssl_opts(self, sslopts):
        if sslopts in [True, False, None, {}]:
            return sslopts
        try:
            if "context" in sslopts:
                return sslopts["context"]
            ssl_version = sslopts.get("ssl_version")
            if ssl_version is None:
                ssl_version = ssl.PROTOCOL_TLS_CLIENT

            context = ssl.SSLContext(ssl_version)

            purpose = ssl.Purpose.SERVER_AUTH

            ca_certs = sslopts.get("ca_certs")

            if ca_certs is not None:
                try:
                    context.load_verify_locations(ca_certs)
                except FileNotFoundError as exc:
                    # FileNotFoundError does not have missing filename info, so adding it below.
                    # since this is the only file path that users can pass in
                    # (`connection_verify` in the EH/SB clients) through opts above.
                    exc.filename = {"ca_certs": ca_certs}
                    raise exc from None
            elif context.verify_mode != ssl.CERT_NONE:
                # load the default system root CA certs.
                context.load_default_certs(purpose=purpose)

            certfile = sslopts.get("certfile")
            keyfile = sslopts.get("keyfile")
            if certfile is not None:
                context.load_cert_chain(certfile, keyfile)

            server_hostname = sslopts.get("server_hostname")
            cert_reqs = sslopts.get("cert_reqs", ssl.CERT_REQUIRED)
            if cert_reqs == ssl.CERT_NONE and server_hostname is None:
                context.check_hostname = False
                context.verify_mode = cert_reqs

            return context
        except TypeError:
            raise TypeError("SSL configuration must be a dictionary, or the value True.") from None


class AsyncTransport(AsyncTransportMixin):  # pylint: disable=too-many-instance-attributes
    """Common superclass for TCP and SSL transports."""

    def __init__(
        self,
        host,
        *,
        port=AMQP_PORT,
        ssl_opts=False,
        socket_settings=None,
        raise_on_initial_eintr=True,
        use_tls: bool = True,
        **kwargs,
    ):
        self.connected = False
        self.sock = None
        self.reader = None
        self.writer = None
        self.raise_on_initial_eintr = raise_on_initial_eintr
        self._read_buffer = BytesIO()
        self.host, self.port = to_host_port(host, port)
        self.host = kwargs.get("custom_endpoint") or self.host
        self.port = kwargs.get("custom_port") or self.port
        self.socket_settings = socket_settings
        self.socket_lock = asyncio.Lock()
        self.sslopts = ssl_opts
        self.network_trace_params = kwargs.get("network_trace_params")
        self._use_tls = use_tls

    async def connect(self):
        try:
            # are we already connected?
            if self.connected:
                return
            try:
                # Building ssl opts here instead of constructor, so that invalid cert error is raised
                # when client is connecting, rather then during creation. For uamqp exception parity.
                self.sslopts = self._build_ssl_opts(self.sslopts)
            except FileNotFoundError as exc:
                # FileNotFoundError does not have missing filename info, so adding it below.
                # Assuming that this must be ca_certs, since this is the only file path that
                # users can pass in (`connection_verify` in the EH/SB clients) through sslopts above.
                # For uamqp exception parity. Remove later when resolving issue #27128.
                exc.filename = self.sslopts
                raise exc
            self.reader, self.writer = await asyncio.open_connection(
                host=self.host,
                port=self.port,
                ssl=self.sslopts if self._use_tls else None,
                family=socket.AF_UNSPEC,
                proto=SOL_TCP,
                server_hostname=self.host if self._use_tls else None,
            )
            self.connected = True
            sock = self.writer.transport.get_extra_info("socket")
            if sock:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                self._set_socket_options(sock, self.socket_settings)

        except (OSError, IOError, SSLError) as e:
            _LOGGER.info("Transport connect failed: %r", e, extra=self.network_trace_params)
            # if not fully connected, close socket, and reraise error
            if self.writer and not self.connected:
                self.writer.close()
                await self.writer.wait_closed()
                self.connected = False
            raise

    def _get_tcp_socket_defaults(self, sock):
        tcp_opts = {}
        for opt in KNOWN_TCP_OPTS:
            enum = None
            if opt == "TCP_USER_TIMEOUT":
                try:
                    from socket import TCP_USER_TIMEOUT as enum
                except ImportError:
                    # should be in Python 3.6+ on Linux.
                    enum = 18
            elif hasattr(socket, opt):
                enum = getattr(socket, opt)

            if enum:
                if opt in DEFAULT_SOCKET_SETTINGS:
                    tcp_opts[enum] = DEFAULT_SOCKET_SETTINGS[opt]
                elif hasattr(socket, opt):
                    tcp_opts[enum] = sock.getsockopt(SOL_TCP, getattr(socket, opt))
        return tcp_opts

    def _set_socket_options(self, sock, socket_settings):
        if socket_settings:
            tcp_opts = self._get_tcp_socket_defaults(sock)
            tcp_opts.update(socket_settings)
            for opt, val in tcp_opts.items():
                sock.setsockopt(SOL_TCP, opt, val)

    async def _read(
        self,
        toread,
        initial=False,
        buffer=None,
        _errnos=(errno.ENOENT, errno.EAGAIN, errno.EINTR),
    ):
        # According to SSL_read(3), it can at most return 16kb of data.
        # Thus, we use an internal read buffer like TCPTransport._read
        # to get the exact number of bytes wanted.
        length = 0
        view = buffer or memoryview(bytearray(toread))
        nbytes = self._read_buffer.readinto(view)
        toread -= nbytes
        length += nbytes
        try:
            while toread:
                try:
                    view[nbytes : nbytes + toread] = await self.reader.readexactly(toread)
                    nbytes = toread
                except AttributeError:
                    # This means that close() was called concurrently
                    # self.reader has been set to None.
                    raise IOError("Connection has already been closed") from None
                except asyncio.IncompleteReadError as exc:
                    pbytes = len(exc.partial)
                    view[nbytes : nbytes + pbytes] = exc.partial
                    nbytes = pbytes
                except socket.error as exc:
                    # ssl.sock.read may cause a SSLerror without errno
                    # http://bugs.python.org/issue10272
                    if isinstance(exc, SSLError) and "timed out" in str(exc):
                        raise socket.timeout()
                    # errno 110 is equivalent to ETIMEDOUT on linux non blocking sockets, when a keep alive is set,
                    # and is set when the connection to the server doesnt succeed
                    # https://man7.org/linux/man-pages/man7/tcp.7.html.
                    # This behavior is linux specific and only on async. sync Linux & async/sync Windows & Mac raised
                    # ConnectionAborted or ConnectionReset errors which properly end up in a retry loop.
                    if exc.errno in [110]:
                        raise ConnectionAbortedError("The connection was closed abruptly.") from exc
                    # ssl.sock.read may cause ENOENT if the
                    # operation couldn't be performed (Issue celery#1414).
                    if exc.errno in _errnos:
                        if initial and self.raise_on_initial_eintr:
                            raise socket.timeout()
                        continue
                    raise
                if not nbytes:
                    raise IOError("Server unexpectedly closed connection")

                length += nbytes
                toread -= nbytes
        except:  # noqa
            self._read_buffer = BytesIO(view[:length])
            raise
        return view

    async def _write(self, s):
        """Write a string out to the SSL socket fully.
        :param str s: The string to write.
        """
        try:
            self.writer.write(s)
            await self.writer.drain()
        except AttributeError:
            raise IOError("Connection has already been closed") from None

    async def close(self):
        async with self.socket_lock:
            try:
                if self.writer is not None:
                    # Closing the writer closes the underlying socket.
                    self.writer.close()
                    if self.sslopts:
                        # see issue: https://github.com/encode/httpx/issues/914
                        await asyncio.sleep(0)
                        self.writer.transport.abort()
                    await self.writer.wait_closed()
            except Exception as e:  # pylint: disable=broad-except
                # Sometimes SSL raises APPLICATION_DATA_AFTER_CLOSE_NOTIFY here on close.
                _LOGGER.debug("Error shutting down socket: %r", e, extra=self.network_trace_params)
            self.writer, self.reader = None, None
        self.connected = False

    async def negotiate(self):
        if not self.sslopts:
            return
        await self.write(TLS_HEADER_FRAME)
        _, returned_header = await self.receive_frame(verify_frame_type=None)
        if returned_header[1] == TLS_HEADER_FRAME:
            raise ValueError(
                f"""Mismatching TLS header protocol. Expected: {TLS_HEADER_FRAME!r},"""
                """received: {returned_header[1]!r}"""
            )


class WebSocketTransportAsync(AsyncTransportMixin):  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        host,
        *,
        port=WEBSOCKET_PORT,
        socket_timeout=CONNECT_TIMEOUT,
        ssl_opts=None,
        use_tls: bool = True,
        **kwargs,
    ):
        self._read_buffer = BytesIO()
        self.socket_lock = asyncio.Lock()
        self.sslopts = ssl_opts if isinstance(ssl_opts, dict) else None
        self.socket_timeout = socket_timeout
        self._custom_endpoint = kwargs.get("custom_endpoint")
        self.host, self.port = to_host_port(host, port)
        self.sock = None
        self.session = None
        self._http_proxy = kwargs.get("http_proxy", None)
        self.connected = False
        self.network_trace_params = kwargs.get("network_trace_params")
        self._use_tls = use_tls

    async def connect(self):
        self.sslopts = self._build_ssl_opts(self.sslopts)
        username, password = None, None
        http_proxy_host, http_proxy_port = None, None
        http_proxy_auth = None

        if self._http_proxy:
            http_proxy_host = self._http_proxy["proxy_hostname"]
            http_proxy_port = self._http_proxy["proxy_port"]
            if http_proxy_host and http_proxy_port:
                http_proxy_host = f"{http_proxy_host}:{http_proxy_port}"
            username = self._http_proxy.get("username", None)
            password = self._http_proxy.get("password", None)

        try:
            from aiohttp import (  # pylint: disable=networking-import-outside-azure-core-transport
                ClientSession,
                ClientConnectorError,
                ClientWSTimeout,
            )
            from urllib.parse import urlsplit
        except ImportError:
            raise ImportError("Please install aiohttp library to use async websocket transport.") from None
        if username or password:
            from aiohttp import BasicAuth  # pylint: disable=networking-import-outside-azure-core-transport

            http_proxy_auth = BasicAuth(login=username, password=password)

        self.session = ClientSession()
        if self._custom_endpoint:
            url = f"wss://{self._custom_endpoint}" if self._use_tls else f"ws://{self._custom_endpoint}"
        else:
            url = f"wss://{self.host}" if self._use_tls else f"ws://{self.host}"
            parsed_url = urlsplit(url)
            url = f"{parsed_url.scheme}://{parsed_url.netloc}:{self.port}{parsed_url.path}"

        try:
            # Enabling heartbeat that sends a ping message every n seconds and waits for pong response.
            # if pong response is not received then close connection. This raises an error when trying
            # to communicate with the websocket which is no longer active.
            # We are waiting a bug fix in aiohttp for these 2 bugs where aiohttp ws might hang on network disconnect
            # and the heartbeat mechanism helps mitigate these two.
            # https://github.com/aio-libs/aiohttp/pull/5860
            # https://github.com/aio-libs/aiohttp/issues/2309

            self.sock = await self.session.ws_connect(
                url=url,
                timeout=ClientWSTimeout(ws_close=self.socket_timeout),
                protocols=[AMQP_WS_SUBPROTOCOL],
                autoclose=False,
                proxy=http_proxy_host,
                proxy_auth=http_proxy_auth,
                ssl=self.sslopts if self._use_tls else None,
                heartbeat=DEFAULT_WEBSOCKET_HEARTBEAT_SECONDS,
            )
        except ClientConnectorError as exc:
            _LOGGER.info("Websocket connect failed: %r", exc, extra=self.network_trace_params)
            if self._custom_endpoint:
                raise AuthenticationException(
                    ErrorCondition.ClientError,
                    description="Failed to authenticate the connection due to exception: " + str(exc),
                    error=exc,
                ) from exc
            raise ConnectionError("Failed to establish websocket connection: " + str(exc)) from exc
        self.connected = True

    async def _read(self, toread, buffer=None, **kwargs):  # pylint: disable=unused-argument
        """Read exactly n bytes from the peer.

        :param int toread: The number of bytes to read.
        :param bytearray or None buffer: The buffer to read into. If not specified, a new buffer is allocated.
        :return: The buffer of bytes read.
        :rtype: bytearray
        """
        length = 0
        view = buffer or memoryview(bytearray(toread))
        nbytes = self._read_buffer.readinto(view)
        length += nbytes
        toread -= nbytes
        try:
            try:
                while toread:
                    data = await self.sock.receive_bytes()
                    read_length = len(data)
                    if read_length <= toread:
                        view[length : length + read_length] = data
                        toread -= read_length
                        length += read_length
                    else:
                        view[length : length + toread] = data[0:toread]
                        self._read_buffer = BytesIO(data[toread:])
                        toread = 0
                return view
            except TypeError as te:
                # aiohttp websocket raises TypeError when a websocket disconnects, as it ends up
                # reading None over the wire and cant convert to bytes.
                raise ConnectionError("Websocket disconnected: %r" % te) from None
        except:
            self._read_buffer = BytesIO(view[:length])
            raise

    async def close(self):
        """Do any preliminary work in shutting down the connection."""
        async with self.socket_lock:
            await self.sock.close()
            await self.session.close()
            self.connected = False

    async def _write(self, s):
        """Completely write a string (byte array) to the peer.
        ABNF, OPCODE_BINARY = 0x2
        See http://tools.ietf.org/html/rfc5234
        http://tools.ietf.org/html/rfc6455#section-5.2

        :param str s: The string to write.
        """
        await self.sock.send_bytes(s)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub with checkpoint store doing checkpoint
by every fixed time interval asynchronously.
In the `receive` method of `EventHubConsumerClient`:
If no partition id is specified, the checkpoint_store are used for load-balance and checkpoint.
If partition id is specified, the checkpoint_store can only be used for checkpoint.
"""
from typing import Dict
import asyncio
import os
import time
from azure.eventhub.aio import EventHubConsumerClient
from azure.eventhub.extensions.checkpointstoreblobaio import BlobCheckpointStore
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]

storage_account_name = os.environ["AZURE_STORAGE_ACCOUNT"]
protocol = os.environ.get("PROTOCOL", "https")
suffix = os.environ.get("ACCOUNT_URL_SUFFIX", "core.windows.net")
BLOB_ACCOUNT_URL = f"{protocol}://{storage_account_name}.blob.{suffix}"
BLOB_CONTAINER_NAME = "your-blob-container-name"  # Please make sure the blob container resource exists.

partition_last_checkpoint_time: Dict[str, float] = {}
checkpoint_time_interval = 15


async def on_event(partition_context, event):
    # Put your code here.
    p_id = partition_context.partition_id
    print("Received event from partition: {}.".format(p_id))
    now_time = time.time()
    p_id = partition_context.partition_id
    last_checkpoint_time = partition_last_checkpoint_time.get(p_id)
    if last_checkpoint_time is None or (now_time - last_checkpoint_time) >= checkpoint_time_interval:
        await partition_context.update_checkpoint(event)
        partition_last_checkpoint_time[p_id] = now_time


async def receive(client):
    """
    Without specifying partition_id, the receive will try to receive events from all partitions and if provided with
    a checkpoint store, the client will load-balance partition assignment with other EventHubConsumerClient instances
    which also try to receive events from all partitions and use the same storage resource.
    """
    await client.receive(
        on_event=on_event,
        starting_position="-1",  # "-1" is from the beginning of the partition.
    )
    # With specified partition_id, load-balance will be disabled, for example:
    # await client.receive(on_event=on_event, partition_id='0'))


async def main():
    checkpoint_store = BlobCheckpointStore(
        blob_account_url=BLOB_ACCOUNT_URL, container_name=BLOB_CONTAINER_NAME, credential=DefaultAzureCredential()
    )
    client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        consumer_group="$Default",
        checkpoint_store=checkpoint_store,  # For load-balancing and checkpoint. Leave None for no load-balancing.
    )
    async with client:
        await receive(client)


if __name__ == "__main__":
    asyncio.run(main())
